# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Function for energy calibration
"""

from uncertainties import ufloat

import hdtv.util
import hdtv.ui
from hdtv.database import IAEALibraries, DDEPLibraries


def SearchNuclide(nuclide, database):
    """
    Searches for information about the nuclide in different databases.
    """
    if database == "IAEA":
        data = IAEALibraries.SearchNuclide(nuclide)
    elif database == "DDEP":
        data = DDEPLibraries.SearchNuclide(nuclide)
    else:
        try:
            data = IAEALibraries.SearchNuclide(nuclide)
        except BaseException:
            data = DDEPLibraries.SearchNuclide(nuclide)
    return data


def TableOfNuclide(data):
    """
    Creates a table of the given data.
    """
    result_header = " ".join(
        [
            "\n",
            "Nuclide:",
            data["nuclide"],
            "\n",
            "Halflife:",
            str(data["halflife"]),
            "\n",
            "Reference:",
            data["reference"],
            "\n",
        ]
    )

    table = hdtv.util.Table(
        data=data["transitions"],
        keys=["energy", "intensity"],
        extra_header=result_header,
        sortBy=None,
        ignoreEmptyCols=False,
    )

    hdtv.ui.msg(html=str(table))


def MatchPeaksAndEnergies(peaks, energies, sigma):
    """
    Combines Peaks with the right energies from the table (with searchEnergie).
    """
    gradient = []  # list of all gradients energy/PeakPosition
    pair = []  # list of all possible pairs energy, Peak
    accordanceCount = []

    # error message if there are no given peaks
    if peaks == []:
        raise hdtv.cmdline.HDTVCommandError("You must fit at least one peak.")

    # saves all pairs and gradients in lists
    for peak in peaks:
        for energy in energies:
            gradient.append(1.0 * energy / peak)
            pair.append([peak, energy])
            accordanceCount.append(0)

    NumberHighestAccordance = 0
    bestAccordance = 0  # gradient with best accordance to the others

    # compare all gradients with each other to find the most frequently one
    # (within sigma)
    for i in range(0, len(gradient)):
        for j in range(0, len(gradient)):
            if gradient[j] > gradient[i] - sigma and gradient[j] < gradient[i] + sigma:
                accordanceCount[i] = accordanceCount[i] + 1
                if accordanceCount[i] > NumberHighestAccordance:
                    NumberHighestAccordance = accordanceCount[i]
                    bestAccordance = gradient[i]

    accordance = []  # all pairs with the right gradient will be saved in this list

    for i in range(0, len(gradient)):
        if (
            gradient[i] > bestAccordance - sigma
            and gradient[i] < bestAccordance + sigma
        ):
            for a in accordance:
                if a[0] == pair[i][0] or a[1] == pair[i][1]:
                    hdtv.ui.msg(f"{a} {pair[i]}")
                    hdtv.ui.warning("Some peaks/energies are used more than one time.")
            accordance.append(pair[i])

    if len(accordance) < 4:
        # hdtv.ui.msg(accordance)
        hdtv.ui.warning("Only few (peak,energy) pairs found. Try with a larger --sigma")

    return accordance


def MatchFitsAndTransitions(fits, transitions, sigma=0.5):
    """
    Combines peaks with the right intensities.
    """
    return [
        {"fit": fit, "transition": transition}
        for fit in fits
        for transition in transitions
        if abs(transition["energy"] - fit.ExtractParams()[0][0]["pos"]) <= sigma
    ]



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import os

import pytest

from tests.helpers.utils import redirect_stdout, hdtvcmd

from hdtv.util import monkey_patch_ui

monkey_patch_ui()

import hdtv.cmdline
import hdtv.options
import hdtv.session

import __main__

try:
    __main__.spectra = hdtv.session.Session()
except RuntimeError:
    pass


from hdtv.plugins.specInterface import spec_interface
import hdtv.plugins.calInterface
import hdtv.plugins.fitInterface
import hdtv.plugins.peakfinder
import hdtv.plugins.fitmap

spectra = __main__.spectra

testspectrum = os.path.join(os.path.curdir, "tests", "share", "osiris_bg.spc")


def count_peak_positions():
    count = 0
    spec = __main__.spectra.dict[__main__.spectra.activeID]
    for fit in spec.dict.values():
        for peak in fit.peaks:
            if "pos_lit" in peak.extras:
                count += 1
    return count


@pytest.fixture(autouse=True)
def prepare():
    spectra.Clear()
    yield
    spectra.Clear()


def test_cmd_fit_position():
    spec_interface.tv.specIf.LoadSpectra(testspectrum, None)
    hdtvcmd("fit peakfind -a -t 0.05")
    f, ferr = hdtvcmd(
        "fit position assign 10.0 1173.228(3) 12.0 1332.492(4)",
        "fit position erase 10.0 12.0",
    )
    assert ferr == ""


def test_cmd_fit_position_map():
    spec_interface.tv.specIf.LoadSpectra(testspectrum, None)
    hdtvcmd("fit peakfind -a -t 0.002")
    f, ferr = hdtvcmd("fit position map tests/share/osiris_bg.map")
    assert ferr == ""
    assert "Mapped 3 energies to peaks" in f


def test_cmd_fit_position_map_tolerance():
    spec_interface.tv.specIf.LoadSpectra(testspectrum, None)
    hdtvcmd("fit peakfind -a -t 0.002")
    f, ferr = hdtvcmd("fit position map -t 8 tests/share/osiris_bg.map")
    assert ferr == ""
    assert "Mapped 2 energies to peaks" in f


def test_cmd_fit_position_map_overwrite():
    test_cmd_fit_position_map()
    assert count_peak_positions() == 3
    f, ferr = hdtvcmd("fit position map -t 2 -o tests/share/osiris_bg.map")
    assert ferr == ""
    assert "Mapped 0 energies to peaks" in f
    assert count_peak_positions() == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

# -------------------------------------------------------------------------
# Program to test handling of spectra with explicit errors in hdtv
# -------------------------------------------------------------------------

# Spectra that are directly recorded usually have errors \sqrt{n} for each bin,
# where n is the number of counts in the bin. However, after scaling, this is
# no longer true. Thus, scaled spectra need to specify the bin errors
# explicitly.
# This program can be used to test the handling of spectra with explicit
# errors in hdtv.

import ROOT
import peakgen

spec = peakgen.Spectrum(1024, -0.5, 1023.5)
spec.func.background = peakgen.PolyBg([10.0])
spec.func.peaks.append(peakgen.TheuerkaufPeak(515, 3000.0, 10.0))

rfile = ROOT.TFile("errtest.root", "RECREATE")

nsamples = int(3e3)

# unscaled histogram
h1 = spec.GetSampledHist(nsamples, "unscaled")

# scaled histogram WITHOUT correct errors
h2 = spec.GetSampledHist(nsamples, "scaled_no_error")
h2.Scale(0.1)

# scaled histogram with correct errors
h3 = spec.GetSampledHist(nsamples, "scaled_with_error")
h3.Sumw2()
h3.Scale(0.1)

# Dump histograms to text files
peakgen.write_hist(h1, "unscaled.txt", include_x=True, include_err=True)
peakgen.write_hist(h2, "scaled_no_error.txt", include_x=True, include_err=True)
peakgen.write_hist(h3, "scaled_with_error.txt", include_x=True, include_err=True)

# Write ROOT file (histogram have been added automatically)
rfile.Write()
rfile.Close()



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Plugin implementing some useful hdtv commands for navigation the file system
his includes a ls-like command, by special request from R. Schulze :)
"""

import os
import stat
import glob

import hdtv.tabformat
import hdtv.cmdline
import hdtv.ui
import hdtv.util


def ls(args):
    """
    this function prints an output similar to that of the ``ls'' program
    """
    if len(args) > 0:
        pattern = os.path.expanduser(args[0])
    else:
        pattern = "*"

    dirlist = []
    for fname in glob.glob(pattern):
        # For broken symlinks, os.stat may fail.
        # We simply add them to the list, without
        # knowing more about them.
        try:
            mode = os.stat(fname)[stat.ST_MODE]
            if stat.S_ISDIR(mode):
                dirlist.append(fname + "/")
            else:
                dirlist.append(fname)
        except OSError:
            dirlist.append(fname)

    dirlist = sorted(dirlist, key=hdtv.util.natural_sort_key)
    hdtv.tabformat.tabformat(dirlist)


def cd(args):
    """
    change current working directory
    """
    if len(args) == 0:
        path = os.path.expanduser("~")
    elif args[0] == "-":
        path = os.environ["OLDPWD"]
    else:
        path = os.path.expanduser(args[0])
    try:
        os.environ["OLDPWD"] = os.getcwd()
        os.chdir(path)
        hdtv.ui.msg(path)
    except OSError as msg:
        hdtv.ui.msg(msg)


def pwd(args):
    """
    print name of current/working directory
    """
    hdtv.ui.msg(os.getcwd())


hdtv.cmdline.AddCommand("ls", ls, level=2, maxargs=1, dirargs=True)
hdtv.cmdline.AddCommand("cd", cd, level=2, maxargs=1, dirargs=True)
hdtv.cmdline.AddCommand("pwd", pwd, level=2, nargs=0)

hdtv.ui.debug("Loaded ls plugin")



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import hdtv.color

from hdtv.weakref_proxy import weakref
from hdtv.drawable import Drawable
from hdtv.util import LockViewport


class Cut(Drawable):
    def __init__(self, color=None, cal=None):
        self.regionMarkers = hdtv.marker.MarkerCollection(
            "X", paired=True, maxnum=1, color=hdtv.color.cut
        )
        self.bgMarkers = hdtv.marker.MarkerCollection(
            "X", paired=True, color=hdtv.color.cut, connecttop=False
        )
        Drawable.__init__(self, color, cal)
        self.spec = None
        self.axis = None  # <- keep this last (needed for __setattr__)

    # delegate everything to the markers
    def __setattr__(self, name, value):
        if name == "cal":
            value = hdtv.cal.MakeCalibration(value)
        if hasattr(self, "axis"):
            self.regionMarkers.__setattr__(name, value)
            self.bgMarkers.__setattr__(name, value)
        Drawable.__setattr__(self, name, value)

    # color property
    def _set_color(self, color):
        # we only need the passive color for fits
        self._passiveColor = hdtv.color.Highlight(color, active=False)
        with LockViewport(self.viewport):
            self.regionMarkers.color = color
            self.bgMarkers.color = color
            if hasattr(self, "spec") and self.spec is not None:
                self.spec.color = color

    def _get_color(self):
        return self._passiveColor

    color = property(_get_color, _set_color)

    def SetMarker(self, mtype, pos):
        if mtype == "":
            if self.regionMarkers.IsFull():
                mtype = "bg"
            else:
                mtype = "region"
        markers = getattr(self, "%sMarkers" % mtype)
        markers.SetMarker(pos)

    def RemoveMarker(self, mtype, pos):
        markers = getattr(self, "%sMarkers" % mtype)
        markers.RemoveNearest(pos)

    def ExecuteCut(self, matrix, axis):
        if self.regionMarkers.IsPending() or len(self.regionMarkers) == 0:
            return None
        self.matrix = matrix
        self.axis = axis
        spec = self.matrix.ExecuteCut(self)
        self.spec = weakref(spec)
        return spec

    def __copy__(self):
        """
        copies marker of this cut
        """
        new = Cut()
        for marker in self.regionMarkers:
            new.regionMarkers.SetMarker(marker.p1.pos_cal)
            if marker.p2:
                new.regionMarkers.SetMarker(marker.p2.pos_cal)
        for marker in self.bgMarkers:
            new.bgMarkers.SetMarker(marker.p1.pos_cal)
            if marker.p2:
                new.bgMarkers.SetMarker(marker.p2.pos_cal)
        return new

    def Draw(self, viewport):
        if not viewport:
            return
        self.viewport = viewport
        self.regionMarkers.Draw(viewport)
        self.bgMarkers.Draw(viewport)

    def Hide(self):
        if not self.viewport:
            return
        with LockViewport(self.viewport):
            self.regionMarkers.Hide()
            self.bgMarkers.Hide()

    def Show(self):
        if not self.viewport:
            return
        with LockViewport(self.viewport):
            self.regionMarkers.Show()
            self.bgMarkers.Show()

    def Refresh(self):
        # FIXME
        if self.matrix:
            self.matrix.ExecuteCut(self, self.axis)



from .exponential import BackgroundModelExponential
from .polynomial import BackgroundModelPolynomial
from .interpolation import BackgroundModelInterpolation

# dictionary of available background models
BackgroundModels = dict()
BackgroundModels["exponential"] = BackgroundModelExponential
BackgroundModels["polynomial"] = BackgroundModelPolynomial
BackgroundModels["interpolation"] = BackgroundModelInterpolation



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2019  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Test whether the calbin algorithm conserves the statistical properties of the spectrum by fitting a peak before and after calbinning.
"""

import os

from math import sqrt

import pytest
import xml.etree.ElementTree as ET

from tests.helpers.utils import hdtvcmd, isclose
from tests.helpers.create_test_spectrum import ArtificialSpec, ArtificialSpecProp
from tests.helpers.fixtures import temp_file

from hdtv.util import monkey_patch_ui

monkey_patch_ui()

import hdtv.cmdline
import hdtv.options
import hdtv.session


import __main__

try:
    __main__.spectra = hdtv.session.Session()
except RuntimeError:
    pass

from hdtv.plugins.specInterface import spec_interface
from hdtv.plugins.fitInterface import fit_interface
from hdtv.plugins.fitlist import fitxml

spectra = __main__.spectra

N_SIGMA = 3  # Determines how many standard deviations (sigma) a fitted quantity
# may deviate from the ideal value.
# Sigma is provided by the fitting algorithm .
UNCERTAINTY_RELATIVE_TOLERANCE = 0.2  # Determines how large the relative deviation, after calbinning, of the fit uncertainties from the original uncertainty may be.
WRITE_BATCHFILE = False  # Determines whether the batch files of the test fits should be written to file.


@pytest.fixture
def test_spectrum(tmp_path):
    ts = ArtificialSpec(path=tmp_path)
    ts.create()
    return ts


def test_calbin(temp_file, test_spectrum):
    command = ["spectrum get " + test_spectrum.filename]

    step = 2  # Use the third spectrum which has a constant background and Poissonian fluctuations
    # Fit the peak in step 2
    command.append(
        "fit marker background set %f"
        % ((step + test_spectrum.bg_regions[0][0]) * test_spectrum.nbins_per_step)
    )
    command.append(
        "fit marker background set %f"
        % ((step + test_spectrum.bg_regions[0][1]) * test_spectrum.nbins_per_step)
    )
    command.append(
        "fit marker background set %f"
        % ((step + test_spectrum.bg_regions[1][0]) * test_spectrum.nbins_per_step)
    )
    command.append(
        "fit marker background set %f"
        % ((step + test_spectrum.bg_regions[1][1]) * test_spectrum.nbins_per_step)
    )

    command.append(
        "fit marker region set %f"
        % (
            (step + 0.5) * test_spectrum.nbins_per_step
            - 3.0 * test_spectrum.peak_width * test_spectrum.nbins_per_step
        )
    )
    command.append(
        "fit marker region set %f"
        % (
            (step + 0.5) * test_spectrum.nbins_per_step
            + 3.0 * test_spectrum.peak_width * test_spectrum.nbins_per_step
        )
    )
    command.append(
        "fit marker peak set %f" % ((step + 0.5) * test_spectrum.nbins_per_step)
    )

    command.append("fit execute")
    command.append("fit store")
    f, ferr = hdtvcmd(*command)

    if WRITE_BATCHFILE:
        batchfile = os.path.join(os.path.curdir, "test", "share", "calbin.hdtv")
        with open(batchfile, "w") as bfile:
            for c in command:
                bfile.write(c + "\n")

        assert ferr == ""

    # Write the fit result to a temporary XML file
    fitxml.WriteXML(spectra.Get("0").ID, temp_file)
    fitxml.ReadXML(spectra.Get("0").ID, temp_file, refit=True, interactive=False)

    # Parse XML file manually
    tree = ET.parse(temp_file)
    root = tree.getroot()

    # Check the number of fits
    fits = root.findall("fit")
    assert len(fits) == 1

    # Read out the main fitted peak properties (position, volume, width) and their uncertainties
    pos_value_init = float(
        fits[0].find("peak").find("cal").find("pos").find("value").text
    )
    pos_error_init = abs(
        float(fits[0].find("peak").find("cal").find("pos").find("error").text)
    )
    vol_value_init = float(
        fits[0].find("peak").find("cal").find("vol").find("value").text
    )
    vol_error_init = abs(
        float(fits[0].find("peak").find("cal").find("vol").find("error").text)
    )
    width_value_init = float(
        fits[0].find("peak").find("cal").find("width").find("value").text
    )
    width_error_init = abs(
        float(fits[0].find("peak").find("cal").find("width").find("error").text)
    )

    # Calbin the spectrum with standard settings, read in the fitted value again and check whether they have changed
    command = ["spectrum calbin 0 -s 1"]
    command.append("fit execute")
    command.append("fit store")
    f, ferr = hdtvcmd(*command)
    assert ferr == ""

    if WRITE_BATCHFILE:
        batchfile = os.path.join(os.path.curdir, "test", "share", "calbin.hdtv")
        with open(batchfile, "a") as bfile:
            for c in command:
                bfile.write(c + "\n")

        assert ferr == ""

    fitxml.WriteXML(spectra.Get("0").ID, temp_file)
    fitxml.ReadXML(spectra.Get("0").ID, temp_file, refit=False, interactive=False)

    tree = ET.parse(temp_file)
    root = tree.getroot()

    fits = root.findall("fit")
    # It is a little bit unsatisfactory to accumulate the fits multiple times in temp_file, but I found no way to erase the content of temp_file
    # open(temp_file, 'w').close() wouldn't work
    assert len(fits) == 3

    pos_value_1 = float(fits[2].find("peak").find("cal").find("pos").find("value").text)
    pos_error_1 = abs(
        float(fits[2].find("peak").find("cal").find("pos").find("error").text)
    )
    vol_value_1 = float(fits[2].find("peak").find("cal").find("vol").find("value").text)
    vol_error_1 = abs(
        float(fits[2].find("peak").find("cal").find("vol").find("error").text)
    )
    width_value_1 = float(
        fits[2].find("peak").find("cal").find("width").find("value").text
    )
    width_error_1 = abs(
        float(fits[2].find("peak").find("cal").find("width").find("error").text)
    )

    # For the fit of the position, allow it to deviate by N_SIGMA times the combined standard deviations of the fitted positions PLUS the width of a single bin.
    assert isclose(
        pos_value_init - pos_value_1,
        0.0,
        abs_tol=N_SIGMA
        * sqrt(pos_error_init * pos_error_init + pos_error_1 * pos_error_1)
        + 1.0,
    )
    assert isclose(
        vol_value_init - vol_value_1,
        0.0,
        abs_tol=N_SIGMA
        * sqrt(vol_error_init * vol_error_init + vol_error_1 * vol_error_1),
    )
    assert isclose(
        width_value_init - width_value_1,
        0.0,
        abs_tol=N_SIGMA
        * sqrt(width_error_init * width_error_init + width_error_1 * width_error_1),
    )

    # Calbin the spectrum with a factor of 2, read in the fitted value again and check whether they have changed
    command = ["spectrum calbin 0 -b 2 -s 0"]
    command.append("fit execute")
    command.append("fit store")
    f, ferr = hdtvcmd(*command)
    assert ferr == ""

    if WRITE_BATCHFILE:
        batchfile = os.path.join(os.path.curdir, "test", "share", "calbin.hdtv")
        with open(batchfile, "a") as bfile:
            for c in command:
                bfile.write(c + "\n")

        assert ferr == ""

    fitxml.WriteXML(spectra.Get("0").ID, temp_file)
    fitxml.ReadXML(spectra.Get("0").ID, temp_file, refit=False, interactive=False)

    tree = ET.parse(temp_file)
    root = tree.getroot()

    fits = root.findall("fit")
    assert len(fits) == 7

    pos_value_2 = float(fits[6].find("peak").find("cal").find("pos").find("value").text)
    pos_error_2 = abs(
        float(fits[6].find("peak").find("cal").find("pos").find("error").text)
    )
    vol_value_2 = float(fits[6].find("peak").find("cal").find("vol").find("value").text)
    vol_error_2 = abs(
        float(fits[6].find("peak").find("cal").find("vol").find("error").text)
    )
    width_value_2 = float(
        fits[6].find("peak").find("cal").find("width").find("value").text
    )
    width_error_2 = abs(
        float(fits[6].find("peak").find("cal").find("width").find("error").text)
    )

    assert isclose(
        pos_value_init - pos_value_2,
        0.0,
        abs_tol=N_SIGMA
        * sqrt(pos_error_init * pos_error_init + pos_error_2 * pos_error_2)
        + 2.0,
    )
    assert isclose(
        vol_value_init - vol_value_2,
        0.0,
        abs_tol=N_SIGMA
        * sqrt(vol_error_init * vol_error_init + vol_error_2 * vol_error_2),
    )
    assert isclose(
        width_value_init - width_value_2,
        0.0,
        abs_tol=N_SIGMA
        * sqrt(width_error_init * width_error_init + width_error_2 * width_error_2),
    )

    assert isclose(pos_error_init, pos_error_2, rel_tol=0.2)
    assert isclose(vol_error_init, vol_error_2, rel_tol=0.2)
    assert isclose(pos_error_init, pos_error_2, rel_tol=0.2)



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Infrastructure for configuration variables
"""

from html import escape


class Option(object):
    """
    A configuration variable
    """

    def __init__(
        self,
        default=None,
        parse=lambda x: x,
        toStr=lambda x: str(x),
        changeCallback=None,
    ):
        self.defaultValue = default
        self.value = self.defaultValue
        self.parse = parse
        self.ToStr = toStr
        self.ChangeCallback = changeCallback

    def __bool__(self):
        """
        How to convert option to 'bool'
        """
        return bool(self.value)

    def Set(self, value):
        """
        Set the variable to the specified value
        """
        self.value = value
        if self.ChangeCallback:
            self.ChangeCallback(self)

    def ParseAndSet(self, rawValue):
        """
        Parses rawValue and sets the variable to the result
        """
        self.Set(self.parse(rawValue))

    def Get(self):
        """
        Return the value of the variable
        """
        return self.value

    def Reset(self):
        """
        Reset the variable to its default value
        """
        self.Set(self.defaultValue)

    def __str__(self):
        """
        Returns the value as a string
        """
        return self.ToStr(self.value)


class _OptionManager(dict):
    """Manages a set of options"""

    def RegisterOption(self, varname, value):
        """
        Adds a configuration variable
        """
        if varname in self.__dict__:
            msg = "Option {!r} already exists with, refusing to overWrite"
            raise RuntimeError(msg.format(varname))
        self.__dict__[varname] = value

    def Set(self, varname, rawValue):
        """
        Sets the variable specified by varname. Raises a KeyError if it does not exist.
        """
        self.__dict__[varname].ParseAndSet(rawValue)

    def Get(self, varname):
        """
        Gets the value of the variable varname. Raises a KeyError if it does not exist.
        """
        return self.__dict__[varname].Get()

    def Reset(self, varname):
        """
        Resets value of variable varname to default. Raises KeyError if it does not exist.
        """
        self.__dict__[varname].Reset()

    def ResetAll(self):
        """
        Resets value of all variables to default.
        """
        for key in self.__dict__.keys():
            self.__dict__[key].Reset()

    def Show(self, varname):
        """
        Shows the value of the variable varname
        """
        return "<b>{}</b>: {}".format(
            escape(varname), escape(str(self.__dict__[varname]))
        )

    def Str(self):
        """
        Returns all options as a string
        """
        string = ""
        for (k, v) in sorted(self.__dict__.items()):
            string += "<b>{}</b>: {}\n".format(escape(k), escape(str(v)))
        return string


def parse_bool(x):
    """
    Parse boolean options
    """
    if x.lower() == "true":
        return True
    elif x.lower() == "false":
        return False
    else:
        raise ValueError("Valid options: True, False.")


def parse_choices(choices):
    def _parse_choices(x):
        if x in choices:
            return x
        else:
            raise ValueError("Valid options: " + ", ".join(choices) + ".")

    return _parse_choices


OptionManager = _OptionManager()

RegisterOption = OptionManager.RegisterOption
Set = OptionManager.Set
Get = OptionManager.Get
Reset = OptionManager.Reset
ResetAll = OptionManager.ResetAll
Show = OptionManager.Show
Str = OptionManager.Str



#!/usr/bin/env python
from __future__ import print_function

import sys

sys.path.append("/home/braun/projects/hdtv")
import hdtv.specreader
from hdtv.errvalue import ErrValue
import ROOT
import math

hdtv.rootext.fit


def test(bgdeg):
    int_tot = ROOT.HDTV.Fit.TH1Integral(h, 6, 12)

    bg = ROOT.HDTV.Fit.PolyBg(bgdeg)
    bg.AddRegion(-0.1, 5.1)
    bg.AddRegion(12.9, 18.1)
    bg.Fit(h)

    for i in range(0, bg.GetDegree() + 1):
        par = ErrValue(bg.GetFunc().GetParameter(i), bg.GetFunc().GetParError(i))
        print("bg[%d]: %10s" % (i, par.fmt()))

    int_bac = ROOT.HDTV.Fit.BgIntegral(bg, 6, 12, h.GetXaxis())
    int_sub = ROOT.HDTV.Fit.TH1BgsubIntegral(h, bg, 6, 12)

    print("")
    print("type        position           width          volume        skewness")
    for (integral, kind) in zip((int_tot, int_bac, int_sub), ("tot:", "bac:", "sub:")):
        pos = ErrValue(integral.GetMean(), integral.GetMeanError())
        width = ErrValue(integral.GetWidth(), integral.GetWidthError())
        vol = ErrValue(integral.GetIntegral(), integral.GetIntegralError())
        skew = ErrValue(integral.GetRawSkewness(), integral.GetRawSkewnessError())

        print(
            "%s %15s %15s %15s %15s"
            % (kind, pos.fmt(), width.fmt(), vol.fmt(), skew.fmt())
        )


reader = hdtv.specreader.SpecReader()
h = reader.GetSpectrum("test.asc")

h.Sumw2()

print("bgdeg=1")
test(1)

print("")
print("bgdeg=2")
test(2)



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Functions for efficiency, energy calibration
"""

import argparse
import math
from uncertainties import ufloat, ufloat_fromstr

import hdtv.efficiency
import hdtv.cmdline
import hdtv.options
import hdtv.ui
import hdtv.cal
import hdtv.util
from hdtv.fitxml import FitXml
from . import EnergyCalibration


class EffCalIf(object):
    def __init__(self, spectra):

        self.spectra = spectra

        # tv commands
        self.tv = EffCalHDTVInterface(self)
        self.fit = False

    def SetFun(self, spectrumID, name, parameter=None):
        """
        Set efficiency function to use


        Allowed names:  * "wunder" for "Wunder"-Efficiency
                        * "wiedenhoever" for Wiedenhoever-Efficiency
                        * "poly" for polynomial efficiency
                        * "exp" for exponential efficiency
                        * "pow" for power function efficiency
        """
        if parameter is None:
            parameter = list()

        try:
            name = name.lower()
            if name == "wunder":
                self.spectra.dict[spectrumID].effCal = hdtv.efficiency.WunderEff(
                    pars=parameter
                )
            elif name == "wiedenhoever":
                self.spectra.dict[spectrumID].effCal = hdtv.efficiency.WiedenhoeverEff(
                    pars=parameter
                )
            elif name == "poly":
                self.spectra.dict[spectrumID].effCal = hdtv.efficiency.PolyEff(
                    pars=parameter
                )
            elif name == "exp":
                self.spectra.dict[spectrumID].effCal = hdtv.efficiency.ExpEff(
                    pars=parameter
                )
            elif name == "pow":
                self.spectra.dict[spectrumID].effCal = hdtv.efficiency.PowEff(
                    pars=parameter
                )
            # elif name == "orthogonal": #to calibrate with Exp
            #    self.spectra.dict[spectrumID].effCal = hdtv.efficiency.ExpEff(pars=parameter)
            #    self.fit = True
            # elif name == "orthogonal_fit": #to fit with othogonal
            #    self.spectra.dict[spectrumID].effCal = hdtv.efficiency.OrthogonalEff(pars=parameter)
            #    self.fit = True
            else:
                raise hdtv.cmdline.HDTVCommandError(
                    "No such efficiency function %s" % name
                )
        except IndexError:
            raise hdtv.cmdline.HDTVCommandError("Invalid spectrum ID %d" % spectrumID)

    def SetPar(self, spectrumID, parameter):
        """
        Set parameter for efficiency function
        """
        try:
            self.spectra.dict[spectrumID].effCal.parameter = parameter
        except IndexError:
            raise hdtv.cmdline.HDTVCommandError("Invalid spectrum ID %d" % spectrumID)
        except AttributeError:
            raise hdtv.cmdline.HDTVCommandError(
                "No efficiency for spectrum ID %d set" % spectrumID
            )

    def Assign(self, todo):
        """
        Assign efficiency for fit
        """
        pass

    def ReadPar(self, spectrumID, filename):
        """
        Load efficiency parameter and covariance from file
        """
        try:
            self.spectra.dict[spectrumID].effCal.loadPar(filename)
        except IndexError:
            raise hdtv.cmdline.HDTVCommandError("Invalid spectrum ID %d" % spectrumID)
        except AttributeError:
            raise hdtv.cmdline.HDTVCommandError(
                "No efficiency for spectrum ID %d set" % spectrumID
            )

    def ReadCov(self, spectrumID, filename):
        """
        Load efficiency parameter and covariance from file
        """
        try:
            self.spectra.dict[spectrumID].effCal.loadCov(filename)
        except IndexError:
            raise hdtv.cmdline.HDTVCommandError("Invalid spectrum ID %d" % spectrumID)
        except AttributeError:
            raise hdtv.cmdline.HDTVCommandError(
                "No efficiency for spectrum ID %d set" % spectrumID
            )

    def WritePar(self, spectrumID, filename):
        """
        Save efficiency parameter
        """
        try:
            self.spectra.dict[spectrumID].effCal.savePar(filename)
        except IndexError:
            raise hdtv.cmdline.HDTVCommandError("Invalid spectrum ID %d" % spectrumID)
        except AttributeError:
            raise hdtv.cmdline.HDTVCommandError(
                "No efficiency for spectrum ID %d set" % spectrumID
            )

    def WriteCov(self, spectrumID, filename):
        """
        Save efficiency parameter
        """
        try:
            self.spectra.dict[spectrumID].effCal.saveCov(filename)
        except IndexError:
            raise hdtv.cmdline.HDTVCommandError("Invalid spectrum ID %d" % spectrumID)
        except AttributeError:
            raise hdtv.cmdline.HDTVCommandError(
                "No efficiency for spectrum ID %d set" % spectrumID
            )

    def List(self, ids=None):
        """
        List currently used efficiencies
        """
        if ids is None:
            ids = self.spectra.ids

        tabledata = list()
        for ID in ids:
            tableline = dict()
            tableline["ID"] = ID
            try:
                tableline["Name"] = self.spectra.dict[ID].effCal.name
                # TODO: Decent formatting
                tableline["Parameter"] = str(self.spectra.dict[ID].effCal.parameter)
            except AttributeError:
                tableline["Name"] = "-"
                tableline["Parameter"] = "-"
            tabledata.append(tableline)

        table = hdtv.util.Table(
            data=tabledata,
            keys=["ID", "Name", "Parameter"],
            sortBy="ID",
            ignoreEmptyCols=False,
        )
        hdtv.ui.msg(html=str(table))

    def Plot(self, spectrumID):
        """
        Plot efficiency
        """
        try:
            self.spectra.dict[spectrumID].effCal.TF1.Draw()
        except AttributeError:
            raise hdtv.cmdline.HDTVCommandError(
                "No efficiency for spectrum ID %d set" % spectrumID
            )

    def Fit(
        self,
        spectrumIDs,
        filename,
        nuclides,
        coefficients,
        sigma,
        show_graph=False,
        fit_panel=False,
        show_table=False,
        source=None,
    ):
        """
        Plot efficiency
        """

        for spectrumID in spectrumIDs:
            if not self.spectra.dict[spectrumID].effCal:
                raise hdtv.cmdline.HDTVCommandError(
                    "Efficiency function not set for spectrum %d, use calibration efficiency set"
                    % spectrumID
                )

        fitValues = hdtv.util.Pairs()
        tabledata = list()

        if filename is not None:  # at the moment you can only fit from one file
            # TODO: maybe it makes more sense to write another method for this
            fitValues.fromFile(filename, sep=" ")  # TODO: separator
            hdtv.ui.msg(spectrumIDs)
            spectrumID = spectrumIDs[0]
        else:  # the spectrum has to be calibrated
            # try:
            if coefficients[0] is None:
                coefficients[0] = 1
                # raise hdtv.cmdline.HDTVCommandError("You have to give a coefficient for the first spectrum.")

            # the first spectrum/nuclide is used to calculate the factor of the
            # other ones
            spectrumID = spectrumIDs[0]
            nuclide = nuclides[0]
            coefficient = coefficients[0]

            matches = self.CalculateEff(spectrumID, nuclide, coefficient, source, sigma)

            # for table option values are saved
            for match in matches:
                tabledata.append(
                    {
                        "Peak": match["fit"].ExtractParams()[0][0]["pos"],
                        "Efficiency": match["efficiency"],
                        "ID": spectrumID,
                        "Nuclide": nuclide,
                        "Intensity": match["transition"]["intensity"],
                        "Vol": match["fit"].ExtractParams()[0][0]["vol"],
                    }
                )

            energies = [match["fit"].ExtractParams()[0][0]["pos"] for match in matches]
            efficiencies = [match["efficiency"] for match in matches]
            fitValues.fromLists(energies, efficiencies)

            # in the region [0;maxEnergy] the relative efficiency will be
            # corrected to the fit
            maxEnergy = energies[-1]

            if len(spectrumIDs) > 1:
                for j in range(1, len(spectrumIDs)):
                    # if the observed spectrum has no coefficient it has to be
                    # corrected
                    if coefficients[j] is None:
                        NewEff = self.EffCorrection(
                            spectrumIDs[0],
                            maxEnergy,
                            spectrumIDs[j],
                            fitValues,
                            nuclides[j],
                            source,
                            sigma,
                        )
                    # if it has its own coefficient only the efficiency has to
                    # be calculated
                    else:
                        NewEff = self.CalculateEff(
                            spectrumIDs[j], nuclides[j], coefficients[j], source, sigma
                        )

                    # all new efficiencies are added to the first ones
                    energies.extend(
                        [match["fit"].ExtractParams()[0][0]["pos"] for match in NewEff]
                    )
                    efficiencies.extend([match["efficiency"] for match in NewEff])
                    for match in NewEff:
                        tabledata.append(
                            {
                                "Peak": match["fit"].ExtractParams()[0][0]["pos"],
                                "Efficiency": match["efficiency"],
                                "ID": spectrumIDs[j],
                                "Nuclide": nuclides[j],
                                "Intensity": match["transition"]["intensity"],
                                "Vol": match["fit"].ExtractParams()[0][0]["vol"],
                            }
                        )

                    fitValues = hdtv.util.Pairs()

                fitValues.fromLists(energies, efficiencies)

        # Call function to do the fit
        if self.fit:
            self.SetFun(spectrumID, "orthogonal_fit")
            self.spectra.dict[spectrumID].effCal.fit(energies, efficiencies)
            # go back to the start condition
            self.SetFun(spectrumID, "orthogonal")
        else:
            try:
                self.spectra.dict[spectrumID].effCal.fit(fitValues, quiet=False)
            except AttributeError:
                raise hdtv.cmdline.HDTVCommandAbort(
                    "No efficiency for spectrum ID %d set", spectrumID
                )
            if show_graph:
                self.spectra.dict[spectrumID].effCal.TGraph.Draw("a*")
                self.spectra.dict[spectrumID].effCal.TF1.Draw("same")

            if fit_panel:
                self.spectra.dict[spectrumID].effCal.TGraph.FitPanel()

        # if table option is called a table will be created
        if show_table:
            table = hdtv.util.Table(
                data=tabledata,
                keys=["ID", "Nuclide", "Peak", "Efficiency", "Intensity", "Vol"],
                sortBy="ID",
                ignoreEmptyCols=False,
            )
            hdtv.ui.msg(html=str(table))

    def CalculateEff(self, spectrumID, nuclide, coefficient, source, sigma):
        """
        Calculates efficiency from a given spectrum and given nuclide.
        """
        fits = list(self.spectra.dict[hdtv.util.ID(spectrumID)].dict.values())
        data = EnergyCalibration.SearchNuclide(nuclide, source)

        # It matches the Peaks and the given energies
        matches = EnergyCalibration.MatchFitsAndTransitions(
            fits, data["transitions"], sigma=sigma
        )
        if not matches:
            raise hdtv.cmdline.HDTVCommandError(
                "There is no match between the fitted peaks and the energies of the nuclide."
            )

        # Calculate efficiency
        for match in matches:
            match["efficiency"] = match["fit"].ExtractParams()[0][0]["vol"] / (
                coefficient * match["transition"]["intensity"]
            )

        return matches

    def EffCorrection(
        self, referenceID, maxEnergy, spectrumID, fitValues, nuclide, source, sigma
    ):
        """
        If there is one spectrum given without coefficient, you have to correct
        it by calculation the missing factor that it fits to the other one.
        """

        # fit the efficiency of the reference spectrum
        self.spectra.dict[referenceID].effCal.fit(fitValues, False)

        # calculate efficiency values for peaks
        matches = self.CalculateEff(spectrumID, nuclide, 1, source, sigma)

        division = 0.0
        amountDivisionError = 0.0
        amountError = 0.0

        # calculates the factor of the nuclide
        # TODO: maybe a iterative function works better
        try:
            for match in matches:
                energy = match["fit"].ExtractParams()[0][0]["pos"]
                if energy <= maxEnergy:
                    functionValue = self.spectra.dict[referenceID].effCal.value(energy)

                    division = match["efficiency"].nominal_value / functionValue
                    amountDivisionError = amountDivisionError + division * (
                        match["efficiency"].std_dev
                    ) ** (-2)
                    amountError = amountError + (match["efficiency"].std_dev) ** (-2)

        except BaseException:
            division = 0.0
            amountDivisionError = 0.0
            amountError = 0.0

            for match in matches:
                energy = match["fit"].ExtractParams()[0][0]["pos"]
                if energy <= maxEnergy:
                    functionValue = self.spectra.dict[referenceID].effCal.value(energy)

                    division = match["efficiency"].nominal_value / functionValue
                    amountDivisionError = amountDivisionError + division
                    amountError = amountError + 1

        if amountError == 0:
            division = 0.0
            amountDivisionError = 0.0
            amountError = 0.0

            for match in matches:
                energy = match["fit"].ExtractParams()[0][0]["pos"]
                functionValue = self.spectra.dict[referenceID].effCal.nominal_value(
                    energy
                )

                division = match["efficiency"].nominal_value / functionValue
                amountDivisionError = amountDivisionError + division
                amountError = amountError + 1

        # the mean value of the divisions is calculated
        factor = amountDivisionError / amountError

        # the corrected efficiency is calculated
        # Efficiency[1] = list(map(lambda x: x / factor, Efficiency[1]))
        for match in matches:
            match["efficiency"] = match["efficiency"] / factor

        return matches


class EffCalHDTVInterface(object):
    def __init__(self, EffCalIf):

        self.effIf = EffCalIf
        self.spectra = EffCalIf.spectra

        self.opt = dict()

        self.opt["eff_fun"] = hdtv.options.Option(
            default="wunder",
            parse=hdtv.options.parse_choices(
                ["wunder", "wiedenhoever", "poly", "exp", "pow"]
            ),
        )  # +orthogonal
        hdtv.options.RegisterOption(
            "calibration.efficiency.function", self.opt["eff_fun"]
        )

        prog = "calibration efficiency set"
        description = "Set efficiency function"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            help="Spectrum ID to set efficiency for",
            action="store",
            default="active",
        )
        parser.add_argument(
            "-p",
            "--parameter",
            help="Parameters for efficiency function",
            action="store",
            default=None,
        )
        parser.add_argument(
            "-f",
            "--file",
            help="Read efficiency from file",
            action="store",
            default=None,
        )
        parser.add_argument(
            "function",
            help="efficiency function for calibration",
            choices=["wunder", "wiedenhoever", "poly", "exp", "pow"],
        )  # +orthogonal
        hdtv.cmdline.AddCommand(prog, self.SetFun, parser=parser)

        prog = "calibration efficiency read parameter"
        description = "Read parameter for efficiency function from file"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            help="Spectrum IDs to read parameters for",
            action="store",
            default="active",
        )
        parser.add_argument(
            "filename",
            metavar="parameter-file",
            help="file with efficiency function parameters",
        )
        hdtv.cmdline.AddCommand(prog, self.ReadPar, parser=parser, fileArgs=True)

        prog = "calibration efficiency read covariance"
        description = "Read covariance matrix of efficiency function from file"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            help="Spectrum IDs to read covariance for",
            action="store",
            default="active",
        )
        parser.add_argument(
            "filename",
            metavar="covariance-file",
            help="file with efficiency covariance matrix",
        )
        hdtv.cmdline.AddCommand(prog, self.ReadCov, parser=parser, fileArgs=True)

        prog = "calibration efficiency write parameter"
        description = "Write parameters of efficiency function from file"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            help="Spectrum ID from which to save parameters",
            action="store",
            default="active",
        )
        parser.add_argument(
            "filename",
            metavar="parameter-file",
            help="output file for efficiency function parameters",
        )
        hdtv.cmdline.AddCommand(prog, self.WritePar, parser=parser, fileArgs=True)

        prog = "calibration efficiency write covariance"
        description = "Write covariance matrix of efficiency function from file"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            help="Spectrum ID from which to save covariance",
            action="store",
            default="active",
        )
        parser.add_argument(
            "filename",
            metavar="covariance-file",
            help="output file for efficiency covariance matrix",
        )
        hdtv.cmdline.AddCommand(prog, self.WriteCov, parser=parser, fileArgs=True)

        prog = "calibration efficiency plot"
        description = "Plot efficiency of spectrum"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "spectrum",
            metavar="spectrum-id",
            help="Spectrum ID to plot efficiency from",
            default="active",
        )
        hdtv.cmdline.AddCommand(prog, self.PlotEff, parser=parser, fileArgs=False)

        prog = "calibration efficiency fit"
        # TODO:better description
        description = "Fit efficiency. For each spectrum a nuclide is necessary and for the first a coefficient or a file."
        parser = hdtv.cmdline.HDTVOptionParser(
            prog=prog,
            description=description,
            epilog="Example: %(prog)s 4,Ra-226,1e7 3,Eu-152 1,Ba-133 -gpt",
        )
        parser.add_argument(
            "-f",
            "--file",
            help="File with energy-efficiency pairs",
            action="store",
            default=None,
        )  # does not really make sense, see method "Fit"
        parser.add_argument(
            "-p",
            "--fit-panel",
            help="Show fit panel",
            action="store_true",
            default=False,
        )
        parser.add_argument(
            "-g",
            "--show-graph",
            help="Show fitted graph",
            action="store_true",
            default=False,
        )
        parser.add_argument(
            "-t",
            "--show-table",
            help="Show table of fitted peaks",
            action="store_true",
            default=False,
        )
        parser.add_argument(
            "-d",
            "--database",
            action="store",
            default="active",
            help="Database from witch the data should be imported.",
        )
        parser.add_argument(
            "-s",
            "--sigma",
            action="store",
            type=float,
            default=0.5,
            help="allowed error of energy-position",
        )
        parser.add_argument(
            "args",
            metavar="spectrumID,nuclide[,factor]",
            help="Spectrum with nuclide and optional factor",
            nargs="+",
        )
        hdtv.cmdline.AddCommand(prog, self.FitEff, parser=parser, fileArgs=False)

        prog = "calibration efficiency list"
        description = "List efficiencies"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "spectrum",
            metavar="spectrum-ids",
            nargs="*",
            help="Spectrum IDs to list",
            default="active",
        )
        hdtv.cmdline.AddCommand(prog, self.ListEff, parser=parser, fileArgs=False)

    def SetFun(self, args):
        """
        set efficiency function
        """

        eff_fun = args.function

        if args.parameter is not None:
            pars = args.parameter.split(",")
            pars = [float(x) for x in pars]
        else:
            pars = None

        if args.file is None:
            parFileName = None
            covFileName = None
        else:
            parFileName = args.file + ".par"
            covFileName = args.file + ".cov"

        try:
            ids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        except ValueError:
            raise hdtv.cmdline.HDTVCommandError("Invalid ID %s" % args.spectrum)

        for ID in ids:
            self.effIf.SetFun(ID, eff_fun, parameter=pars)
            if parFileName is not None:
                self.effIf.ReadPar(ID, parFileName)
            if covFileName is not None:
                self.effIf.ReadCov(ID, covFileName)

    def ReadPar(self, args):
        """
        Read efficiency parameter
        """
        filename = args.filename

        try:
            ids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        except ValueError:
            raise hdtv.cmdline.HDTVCommandError("Invalid ID %s" % args.spectrum)

        for ID in ids:
            self.effIf.ReadPar(ID, filename)

    def ReadCov(self, args):
        """
        Read efficiency covariance
        """
        filename = args.filename

        try:
            ids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        except ValueError:
            raise hdtv.cmdline.HDTVCommandError("Invalid ID %s" % args.spectrum)

        for ID in ids:
            self.effIf.ReadCov(ID, filename)

    def WritePar(self, args):
        """
        Save efficiency parameter
        """
        filename = args.filename

        try:
            ids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        except ValueError:
            raise hdtv.cmdline.HDTVCommandError("Invalid ID %s" % args.spectrum)

        if len(ids) > 1:
            raise hdtv.cmdline.HDTVCommandError(
                "Can only write efficiency parameter of one spectrum"
            )

        for ID in ids:
            self.effIf.WritePar(ID, filename)

    def WriteCov(self, args):
        """
        Write efficiency covariance
        """
        filename = args.filename

        try:
            ids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        except ValueError:
            raise hdtv.cmdline.HDTVCommandError("Invalid ID %s" % args.spectrum)

        if len(ids) > 1:
            raise hdtv.cmdline.HDTVCommandError(
                "Can only write efficiency covariance of one spectrum"
            )

        for ID in ids:
            self.effIf.WriteCov(ID, filename)

    def PlotEff(self, args):
        """
        Plot efficiency
        """
        try:
            ids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        except ValueError:
            raise hdtv.cmdline.HDTVCommandError("Invalid ID %s" % args.spectrum)

        if len(ids) > 1:
            raise hdtv.cmdline.HDTVCommandError(
                "Can only plot efficiency covariance of one spectrum"
            )

        for ID in ids:
            self.effIf.Plot(ID)

    def FitEff(self, args):
        """
        Fit efficiency
        """
        # TODO: error if spectrum is not visible
        ID = []
        nuclides = []
        coefficients = []
        try:
            ids = hdtv.util.ID.ParseIds(args.args, self.spectra)
        except BaseException:
            pass

        if args.file is None:
            for argument in args.args:
                try:
                    argument = argument.split(",")
                    ID.append(argument[0])
                    nuclides.append(argument[1])
                except BaseException:
                    raise hdtv.cmdline.HDTVCommandError(
                        "You have to give at least a 'spectrumID,nuclide'."
                    )

                if len(argument) > 2:
                    try:
                        coefficients.append(float(argument[2]))
                    except BaseException:
                        raise hdtv.cmdline.HDTVCommandError(
                            "Invalid coefficient %s." % str(argument[2])
                        )
                else:
                    coefficients.append(None)
            try:
                ids = hdtv.util.ID.ParseIds(ID, self.spectra)
            except ValueError:
                raise hdtv.cmdline.HDTVCommandError("Invalid ID %s" % ID)
                return

        self.effIf.Fit(
            ids,
            args.file,
            nuclides,
            coefficients,
            args.sigma,
            args.show_graph,
            args.fit_panel,
            args.show_table,
            args.database,
        )

    def ListEff(self, args):
        """
        List efficiencies
        """
        try:
            ids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        except ValueError:
            raise hdtv.cmdline.HDTVCommandError("Invalid ID %s" % args.spectrum)

        self.effIf.List(ids)


class EnergyCalIf(object):
    """
    Interface for energy calibrations
    """

    def __init__(self, spectra):
        self.spectra = spectra

        # tv commands
        self.tv = EnergyCalHDTVInterface(self)

    def CalFromFile(self, fname):
        """
        Read calibration polynom from file

        Allow formats are:
            * One coefficient in each line, starting with p0
            * Coefficients in one line, seperated by space, starting with p0
        """
        f = hdtv.util.TxtFile(fname)
        f.read()

        try:
            calpoly = []
            for line in f.lines:
                l = line.split()
                if len(l) > 1:  # One line cal file
                    for p in l:
                        calpoly.append(float(p))
                    raise StopIteration
                else:
                    if l[0] != "":
                        calpoly.append(float(l[0]))

        except StopIteration:  # end file reading
            pass

        return hdtv.cal.MakeCalibration(calpoly)

    def CalFromPairs(
        self,
        pairs,
        degree=1,
        table=False,
        fit=False,
        residual=False,
        ignore_errors=False,
    ):
        """
        Create calibration from pairs of channel and energy
        """
        fitter = hdtv.cal.CalibrationFitter()
        for p in pairs:
            fitter.AddPair(p[0], p[1])
        fitter.FitCal(degree, ignore_errors=ignore_errors)
        hdtv.ui.msg(html=fitter.ResultStr())
        if table:
            hdtv.ui.msg(html=str(fitter.ResultTable()))
        if fit:
            fitter.DrawCalFit()
        if residual:
            fitter.DrawCalResidual()
        return fitter.calib

    def CalFromFits(
        self,
        fits,
        pairs,
        degree=1,
        table=False,
        fit=False,
        residual=False,
        ignore_errors=False,
    ):
        """
        Create a calibration from pairs of fits and energies
        """
        valid_pairs = hdtv.util.Pairs()
        for p in pairs:
            pid = p[0].minor
            if pid is None:
                pid = 0
            fid = p[0]
            fid.minor = None
            try:
                peak = fits[fid].peaks[pid]
            except (IndexError, KeyError):
                fid.minor = pid
                hdtv.ui.warning("Ignoring invalid peak id %s" % fid)
                continue
            peak.extras["pos_lit"] = p[1]
            valid_pairs.add(peak.pos, p[1])
        return self.CalFromPairs(
            valid_pairs, degree, table, fit, residual, ignore_errors=ignore_errors
        )

    def CalsFromList(self, fname):
        """
        Reads calibrations from a calibration list file. The file has the format
        <specname>: <cal0> <cal1> ...
        The calibrations are written into the calibration dictionary.
        """
        calDict = dict()
        f = hdtv.util.TxtFile(fname)
        f.read()
        for (l, n) in zip(f.lines, f.linos):
            try:
                (k, v) = l.split(":", 1)
                name = k.strip()
                coeff = [float(s) for s in v.split()]
                calDict[name] = hdtv.cal.MakeCalibration(coeff)
            except ValueError:
                hdtv.ui.warning(
                    "Could not parse line %d of file %s: ignored." % (n, fname)
                )
        return calDict

    def CreateCalList(self, calDict):
        """
        Creates a printable list of all calibrations in calDict
        <specname>: <cal0> <cal1> ...
        """
        lines = list()
        names = sorted(calDict.keys())
        for name in names:
            cal = calDict[name]
            lines.append(name + ": " + hdtv.cal.PrintCal(cal))
        return "\n".join(lines)

    def CopyCal(self, source_id, ids):
        """
        Copy a calibration
        """
        cal = self.spectra.dict[source_id].cal
        coeffs = hdtv.cal.GetCoeffs(cal)
        cal = hdtv.cal.MakeCalibration(cal)
        self.spectra.ApplyCalibration(ids, cal)


class EnergyCalHDTVInterface(object):
    def __init__(self, ECalIf):

        self.EnergyCalIf = ECalIf
        self.spectra = ECalIf.spectra

        # calibration commands
        prog = "calibration position set"
        description = "Create calibration from the coefficients p of a polynomial"
        usage = "%(prog)s [-h] [-s SPECTRUM] p0 p1 [p2 ...]"
        parser = hdtv.cmdline.HDTVOptionParser(
            prog=prog, description=description, usage=usage
        )
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="spectrum ids to apply calibration to",
        )
        parser.add_argument("p0", type=float, help=argparse.SUPPRESS)
        parser.add_argument(
            "prest",
            metavar="p1, [p2 ...]",
            nargs="+",
            type=float,
            help=argparse.SUPPRESS,
        )
        hdtv.cmdline.AddCommand(prog, self.CalPosSet, parser=parser)

        prog = "calibration position unset"
        description = "Unset the calibration of spectra"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="spectrum ids to unset calibration",
        )
        hdtv.cmdline.AddCommand(prog, self.CalPosUnset, parser=parser)

        prog = "calibration position copy"
        description = """apply the calibration that is used for the spectrum
                      selected by source-id to the spectra with dest-ids."""
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "sourceid",
            metavar="source-id",
            action="store",
            help="spectrum to copy calibration from",
        )
        parser.add_argument(
            "destids",
            metavar="dest-id",
            action="store",
            help="spectrum to copy calibration to",
            nargs="+",
        )
        hdtv.cmdline.AddCommand(prog, self.CalPosCopy, parser=parser)

        prog = "calibration position enter"
        description = "Fit a calibration polynomial to the energy/channel pairs given. "
        usage = "%(prog)s [OPTIONS] channel0 energy0 [channel1 energy1 ...]"
        parser = hdtv.cmdline.HDTVOptionParser(
            prog=prog,
            description=description,
            usage=usage,
            epilog="""Hint: Specifying degree=0 will fix the linear term at 1.
                   Specify spec=None to only fit the calibration.""",
        )
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="spectrum ids to apply calibration to",
        )
        parser.add_argument(
            "-d",
            "--degree",
            action="store",
            default=1,
            type=int,
            help="degree of calibration polynomial fitted [default: %(default)s]",
        )
        parser.add_argument(
            "-D",
            "--draw-fit",
            action="store_true",
            default=False,
            help="draw fit used to obtain calibration",
        )
        parser.add_argument(
            "-r",
            "--draw-residual",
            action="store_true",
            default=False,
            help="show residual of calibration fit",
        )
        parser.add_argument(
            "-t",
            "--show-table",
            action="store_true",
            default=False,
            help="print table of energies given and energies obtained from fit",
        )
        parser.add_argument(
            "-f",
            "--file",
            action="store",
            default=None,
            help="get channel-energy pairs from file",
        )
        parser.add_argument(
            "-i",
            "--ignore-errors",
            action="store_true",
            default=False,
            help="set all weights to 1 in fit (ignore error bars even if given)",
        )
        parser.add_argument(
            "args",
            metavar="channelN energyN",
            help="channel-energy pairs used for fit",
            nargs=argparse.REMAINDER,
        )
        hdtv.cmdline.AddCommand(
            prog, self.CalPosEnter, level=0, parser=parser, fileargs=True
        )

        prog = "calibration position nuclide"
        description = "Fit a calibration polynomial to a given nuclide."
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-S",
            "--sigma",
            action="store",
            default=0.0001,
            type=float,
            help="allowed error by variation of energy/channel",
        )
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default=None,
            help="spectrum ids to apply calibration to",
            nargs=1,
        )
        parser.add_argument(
            "-d",
            "--database",
            action="store",
            default="active",
            help="Database from witch the data should be imported.",
        )
        parser.add_argument("nuclide", nargs="+", help="nuclide to use for calibration")
        hdtv.cmdline.AddCommand(prog, self.CalPosNuc, parser=parser)

        prog = "nuclide"
        description = "Prints out the Information of the nuclide."
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-d",
            "--database",
            action="store",
            default="active",
            help="Database from witch the data should be imported.",
        )
        parser.add_argument("nuclide", nargs="+", help="nuclide to query")
        hdtv.cmdline.AddCommand(prog, self.Nuc, parser=parser)

        prog = "calibration position read"
        description = """Read the energy calibration from a file containing a single calibration.
        This corresponds to the energy calibration format used by tv:
        The calibration coefficients are all placed in separate lines."""
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="spectrum ids to apply calibration to",
        )
        parser.add_argument(
            "filename", help="file with position calibration parameters"
        )
        hdtv.cmdline.AddCommand(prog, self.CalPosRead, parser=parser, fileargs=True)

        prog = "calibration position assign"
        description = """Calibrate the active spectrum by asigning energies to fitted peaks.
        peaks are specified by their index and the peak number within the peak
        (if number is ommitted the first (and only?) peak is taken)."""
        usage = "%(prog)s [OPTIONS] id0 energy0 [id1 energy1 ...]"
        parser = hdtv.cmdline.HDTVOptionParser(
            prog=prog, description=description, usage=usage
        )
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="spectrum ids to apply calibration to",
        )
        parser.add_argument(
            "-d",
            "--degree",
            action="store",
            default="1",
            help="degree of calibration polynomial fitted [default: %(default)s]",
        )
        parser.add_argument(
            "-f",
            "--show-fit",
            action="store_true",
            default=False,
            help="show fit used to obtain calibration",
        )
        parser.add_argument(
            "-r",
            "--show-residual",
            action="store_true",
            default=False,
            help="show residual of calibration fit",
        )
        parser.add_argument(
            "-t",
            "--show-table",
            action="store_true",
            default=False,
            help="print table of energies given and energies obtained from fit",
        )
        parser.add_argument(
            "-i",
            "--ignore-errors",
            action="store_true",
            default=False,
            help="set all weights to 1 in fit (ignore error bars even if given)",
        )
        parser.add_argument(
            "args",
            metavar="idN energyN",
            help="peak id-energy pair used for calibration",
            nargs=argparse.REMAINDER,
        )
        hdtv.cmdline.AddCommand(
            "calibration position assign", self.CalPosAssign, parser=parser
        )

        prog = "calibration position list"
        description = "List all energy calibrations that are currently loaded."
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        hdtv.cmdline.AddCommand(prog, self.CalPosList, parser=parser)

        prog = "calibration position list write"
        description = "Write all loaded energy calibrations to a file."
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-F",
            "--force",
            action="store_true",
            default=False,
            help="overwrite existing files without asking",
        )
        parser.add_argument(
            "filename", metavar="output-file", help="output file for position list"
        )
        hdtv.cmdline.AddCommand(
            prog, self.CalPosListWrite, parser=parser, fileargs=True
        )

        prog = "calibration position list read"
        description = "Read several energy calibrations from a file."
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "filename",
            metavar="input-file",
            help="input file for calibration position list",
        )
        hdtv.cmdline.AddCommand(prog, self.CalPosListRead, parser=parser, fileargs=True)

        prog = "calibration position list clear"
        description = "Clear all energy calibrations"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        hdtv.cmdline.AddCommand(prog, self.CalPosListClear, parser=parser)

    def Nuc(self, args):
        """
        Returns a table of energies and intensities of the given nuclide.
        """
        for nuclide in args.nuclide:
            data = EnergyCalibration.SearchNuclide(nuclide, args.database)
            EnergyCalibration.TableOfNuclide(data)

    def CalPosSet(self, args):
        """
        Create calibration from the coefficients p of a polynomial
        """
        # parsing command
        cal = [args.p0] + args.prest
        ids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        if not ids:
            hdtv.ui.warning("Nothing to do")
            return
        # do the work
        self.spectra.ApplyCalibration(ids, cal)
        return True

    def CalPosUnset(self, args):
        """
        Unset calibration
        """
        # parsing command
        ids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        if not ids:
            hdtv.ui.warning("Nothing to do")
            return
        self.spectra.ApplyCalibration(ids, None)
        return True

    def CalPosCopy(self, args):
        """
        Copy calibration from one spectrum to others
        """
        self.EnergyCalIf.CopyCal(
            hdtv.util.ID.ParseIds(args.sourceid, self.spectra)[0],
            hdtv.util.ID.ParseIds(args.destids, self.spectra),
        )

    def CalPosEnter(self, args):
        """
        Create calibration from pairs of channel and energy
        """
        # parsing command
        pairs = hdtv.util.Pairs(ufloat_fromstr)
        degree = int(args.degree)
        if args.file is not None:  # Read from file
            pairs.fromFile(args.file)
        else:  # Read from command line
            if len(args.args) % 2 != 0:
                raise hdtv.cmdline.HDTVCommandError("Number of parameters must be even")
            for channel, energy in zip(*[iter(args.args)] * 2):
                pairs.add(channel, energy)
        sids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        # do the work
        cal = self.EnergyCalIf.CalFromPairs(
            pairs,
            degree,
            args.show_table,
            args.draw_fit,
            args.draw_residual,
            ignore_errors=args.ignore_errors,
        )

        if not sids:
            hdtv.ui.msg("calibration: %s" % hdtv.cal.PrintCal(cal))
            return True
        self.spectra.ApplyCalibration(sids, cal)
        return True

    def CalPosNuc(self, args):
        """
        Create a calibration for given nuclide.
        """
        # option spectrum
        spectrumID = []
        try:
            spectrumIDList = list(args.spectrum)
            optionSpectrum = True  # check if option spectrum is called
        except BaseException:
            optionSpectrum = False

        if optionSpectrum:
            # TODO: Use build in -s parsing
            if len(spectrumIDList) > 1:
                # check if string is like '1,2,3 ...'
                for i in range(1, len(spectrumIDList) - 1, 2):
                    if spectrumIDList[i] != ",":
                        raise hdtv.cmdline.HDTVCommandError(
                            "Invalid spectrumID, it has to look like '0,1'"
                        )
            for i in range(
                0, len(spectrumIDList), 2
            ):  # makes a list of all spectrum IDs
                try:
                    # check if spectrumIDs are integers and save them
                    spectrumID.append(int(spectrumIDList[i]))
                except BaseException:
                    raise hdtv.cmdline.HDTVCommandError(
                        "Invalid spectrumID, it has to look like '0,1'"
                    )
        else:
            # check if active spectrum is visible
            if not __main__.spectra.activeID in __main__.spectra.visible:
                raise hdtv.cmdline.HDTVCommandError(
                    "Active spectrum is not visible, no action taken"
                )
            # when no option is called the active spectrum is used
            spectrumID.append(int(self.spectra.activeID))

        Peaks = []

        # position of the fitted peaks saved in peaks
        for ID in spectrumID:
            try:
                fits = self.spectra.dict[hdtv.util.ID(ID)].dict
                for fit in list(fits.values()):
                    Peaks.append(fit.ExtractParams()[0][0]["channel"])
            except BaseException:  # errormessage if there is no spectrum with the given ID
                raise hdtv.cmdline.HDTVCommandError(
                    "Spectrum with ID " + str(ID) + " is not visible, no action taken"
                )

        # finds the right transitions for the given nuclide(s) from table
        nuclei = [
            EnergyCalibration.SearchNuclide(nucl, args.database)
            for nucl in args.nuclide
        ]
        transitions = [transition for n in nuclei for transition in n["transitions"]]
        energies = [t["energy"] for t in transitions]

        # matches the right peaks with the right energy
        Match = EnergyCalibration.MatchPeaksAndEnergies(Peaks, energies, args.sigma)

        # prints all important values
        nuclideStr = " ".join(args.nuclide)
        spectrumIDStr = " ".join([str(i) for i in spectrumID])

        hdtv.ui.msg(
            "Create a calibration for nuclide "
            + nuclideStr
            + " (sigma: "
            + str(args.sigma)
            + ", spectrum"
            + spectrumIDStr
            + ", database "
            + nuclei[0]["reference"]
            + ")"
        )

        # calibration
        degree = 1
        fitter = hdtv.cal.CalibrationFitter()
        for p in Match:  # builds pairs
            fitter.AddPair(p[0], p[1])
        fitter.FitCal(degree, ignore_errors=True)
        hdtv.ui.msg(html=fitter.ResultStr())
        cal = fitter.calib
        for ID in spectrumID:
            self.spectra.ApplyCalibration(ID, cal)  # do the calibration

    def CalPosRead(self, args):
        """
        Read calibration from file
        """
        # parsing command
        sids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        filename = args.filename

        if not sids:
            hdtv.ui.warning("Nothing to do")
            return
        # do the work
        try:
            cal = self.EnergyCalIf.CalFromFile(filename)
            self.spectra.ApplyCalibration(sids, cal)
        except ValueError:
            raise hdtv.cmdline.HDTVCommandError(
                "Malformed calibration parameter file '%s'." % filename
            )
        return True

    def CalPosAssign(self, args):
        """
        Calibrate the active spectrum by assigning energies to fitted peaks

        Peaks are specified by their id and the peak number within the fit.
        Syntax: id.number
        If no number is given, the first peak in the fit is used.
        """
        if self.spectra.activeID is None:
            hdtv.ui.warning("No active spectrum, no action taken.")
            return False
        spec = self.spectra.GetActiveObject()
        # parsing of command
        if len(args.args) % 2 != 0:
            raise hdtv.cmdline.HDTVCommandError("Number of arguments must be even")
        else:
            pairs = hdtv.util.Pairs()
            for peak_id, energy in zip(*[iter(args.args)] * 2):
                ID = hdtv.util.ID.ParseIds(peak_id, spec, only_existent=False)[0]
                value = ufloat_fromstr(energy)
                pairs.add(ID, value)
        sids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        if not sids:
            sids = [self.spectra.activeID]
        degree = int(args.degree)
        # do the work
        cal = self.EnergyCalIf.CalFromFits(
            spec.dict,
            pairs,
            degree,
            table=args.show_table,
            fit=args.show_fit,
            residual=args.show_residual,
            ignore_errors=args.ignore_errors,
        )
        self.spectra.ApplyCalibration(sids, cal)
        return True

    def CalPosList(self, args):
        """
        Print currently known calibration list
        """
        text = self.EnergyCalIf.CreateCalList(self.spectra.caldict)
        hdtv.ui.msg(text)

    def CalPosListWrite(self, args):
        """
        Write calibration list to file
        """
        text = self.EnergyCalIf.CreateCalList(self.spectra.caldict)
        fname = hdtv.util.user_save_file(args.filename, args.force)
        if not fname:
            return
        with open(fname, "w") as calfile:
            calfile.write(text)

    def CalPosListRead(self, args):
        """
        Read calibrations for several spectra from file
        """
        caldict = self.EnergyCalIf.CalsFromList(args.filename)
        if caldict is None:
            return
        # update calcdict of main session
        self.spectra.caldict.update(caldict)
        for name in caldict.keys():
            for sid in self.spectra.dict.keys():
                if self.spectra.dict[sid].name == name:
                    cal = caldict[name]
                    self.spectra.ApplyCalibration([sid], cal)

    def CalPosListClear(self, args):
        """
        Clear list of name <-> calibration pairs
        """
        for name in list(self.spectra.caldict.keys()):
            for sid in self.spectra.dict.keys():
                if self.spectra.dict[sid].name == name:
                    self.spectra.ApplyCalibration([sid], None)
        self.spectra.caldict.clear()


import __main__

eff_cal_interface = EffCalIf(__main__.spectra)
energy_cal_interface = EnergyCalIf(__main__.spectra)
hdtv.cmdline.RegisterInteractive("eff", eff_cal_interface)
hdtv.cmdline.RegisterInteractive("ecal", energy_cal_interface)



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Matrix interface for hdtv
"""

import os

import ROOT
import hdtv.rootext.display

import hdtv.ui
import hdtv.util
import hdtv.cmdline
from hdtv.specreader import SpecReader, SpecReaderError

from hdtv.matrix import Matrix
from hdtv.histogram import MHisto2D


class MatInterface(object):
    def __init__(self, spectra):
        hdtv.ui.debug("Loaded user interface for working with 2-d spectra")

        self.spectra = spectra
        self.window = spectra.window
        self.oldcut = None

        # tv commands
        self.tv = TvMatInterface(self)

        # Register hotkeys
        if self.window:
            self.window.AddHotkey(ROOT.kKey_g, lambda: self.spectra.SetMarker("cut"))
            self.window.AddHotkey(
                [ROOT.kKey_c, ROOT.kKey_g], lambda: self.spectra.SetMarker("cutregion")
            )
            self.window.AddHotkey(
                [ROOT.kKey_c, ROOT.kKey_b], lambda: self.spectra.SetMarker("cutbg")
            )
            self.window.AddHotkey(
                [ROOT.kKey_Minus, ROOT.kKey_c, ROOT.kKey_g],
                lambda: self.spectra.RemoveMarker("cutregion"),
            )
            self.window.AddHotkey(
                [ROOT.kKey_Minus, ROOT.kKey_c, ROOT.kKey_b],
                lambda: self.spectra.RemoveMarker("cutbg"),
            )
            self.window.AddHotkey([ROOT.kKey_Minus, ROOT.kKey_C], self.spectra.ClearCut)
            self.window.AddHotkey([ROOT.kKey_Plus, ROOT.kKey_C], self.spectra.StoreCut)
            self.window.AddHotkey(ROOT.kKey_C, self.spectra.ExecuteCut)
            self.window.AddHotkey(
                [ROOT.kKey_c, ROOT.kKey_s],
                lambda: self.window.EnterEditMode(
                    prompt="Show Cut: ", handler=self._HotkeyShow
                ),
            )
            self.window.AddHotkey(
                [ROOT.kKey_c, ROOT.kKey_a],
                lambda: self.window.EnterEditMode(
                    prompt="Activate Cut: ", handler=self._HotkeyActivate
                ),
            )
            self.window.AddHotkey(
                [ROOT.kKey_c, ROOT.kKey_p], lambda: self._HotkeyShow("PREV")
            )
            self.window.AddHotkey(
                [ROOT.kKey_c, ROOT.kKey_n], lambda: self._HotkeyShow("NEXT")
            )
            self.window.AddHotkey(ROOT.kKey_Tab, self._HotkeySwitch)

    def _HotkeyShow(self, args):
        """
        Show wrapper for use with a Hotkey (internal use)
        """
        spec = self.spectra.GetActiveObject()
        if spec is None and self.window:
            self.window.viewport.SetStatusText("No active spectrum")
            return
        if (not hasattr(spec, "matrix") or spec.matrix is None) and self.window:
            self.window.viewport.SetStatusText("No matrix")
            return
        try:
            ids = hdtv.util.ID.ParseIds(args, spec.matrix)
        except ValueError:
            if self.window:
                self.window.viewport.SetStatusText("Invalid cut identifier: %s" % args)
            return
        spec.ShowObjects(ids)

    def _HotkeyActivate(self, args):
        """
        ActivateObject wrapper for use with a Hotkey (internal use)
        """
        spec = self.spectra.GetActiveObject()
        if spec is None and self.window:
            self.window.viewport.SetStatusText("No active spectrum")
            return
        if (not hasattr(spec, "matrix") or spec.matrix is None) and self.window:
            self.window.viewport.SetStatusText("No matrix")
            return
        try:
            ids = hdtv.util.ID.ParseIds(args, spec.matrix)
        except ValueError:
            if self.window:
                self.window.viewport.SetStatusText("Invalid cut identifier: %s" % args)
            return
        if len(ids) == 1:
            if self.window:
                self.window.viewport.SetStatusText("Activating cut %s" % ids[0])
            self.spectra.ActivateCut(ids[0])
        elif len(ids) == 0:
            if self.window:
                self.window.viewport.SetStatusText("Deactivating cut")
            self.spectra.ActivateCut(None)
        else:
            if self.window:
                self.window.viewport.SetStatusText("Can only activate one cut")

    def _HotkeySwitch(self):
        # FIXME
        spec = self.spectra.GetActiveObject()
        if spec is None:
            raise hdtv.cmdline.HDTVCommandError("There is no active spectrum")
        if not hasattr(spec, "matrix") or spec.matrix is None:
            raise hdtv.cmdline.HDTVCommandError(
                "Active spectrum does not belong to a matrix"
            )
        mat = spec.matrix
        # for a cut, switch to projection
        if spec in mat.specs:
            if spec.axis == "y":
                # show x proj
                ID = self.spectra.Index(mat.xproj)
            else:
                # show y proj
                ID = self.spectra.Index(mat.yproj)
            if ID is None:
                hdtv.ui.warning("Projection is not loaded.")
                return
        # for a projection, cut to last shown cut
        else:
            ID = self.oldcut
            if ID is None:
                # FIXME
                hdtv.ui.warning("No old cut")
                return
        self.spectra.ShowObjects(ID)

    def LoadMatrix(self, fname, sym, ID=None):
        # FIXME: just for testing!
        try:
            histo = MHisto2D(fname, sym)
        except (OSError, SpecReaderError):
            hdtv.ui.warning("Could not load %s" % fname)
            return

        matrix = Matrix(histo, sym, self.spectra.viewport)
        proj = matrix.xproj
        ID = self.spectra.GetFreeID()
        matrix.ID = ID
        matrix.color = hdtv.color.ColorForID(ID.major)
        # load x projection
        sid = self.spectra.Insert(proj, ID=hdtv.util.ID(ID.major, 1000))
        self.spectra.ActivateObject(sid)
        if sym is False:
            # also load y projection
            proj = matrix.yproj
            self.spectra.Insert(proj, ID=hdtv.util.ID(ID.major, 1001))

    def ListMatrix(self, matrix):
        params = ["ID", "stat", "axis", "gates", "bg", "specID"]
        cuts = list()
        count = 0

        for (ID, obj) in matrix.dict.items():

            this = dict()

            stat = str()
            if ID == matrix.activeID:
                stat += "A"
            if ID in matrix.visible:
                stat += "V"

            this["ID"] = ID
            this["stat"] = stat
            this["axis"] = obj.axis

            gates = str()
            for gate in obj.regionMarkers:
                p = sorted([gate.p1.pos_cal, gate.p2.pos_cal])
                gates += " %d - %d " % (p[0], p[1])
            this["gates"] = gates

            bgs = str()
            for bg in obj.bgMarkers:
                bg = sorted([bg.p1.pos_cal, bg.p2.pos_cal])
                bgs += " %d - %d " % (bg[0], bg[1])
            this["bg"] = bgs

            if obj.spec is not None:
                count += 1
                sid = self.spectra.Index(obj.spec)
                stat = str()
                if sid == self.spectra.activeID:
                    stat += "A"
                if sid in self.spectra.visible:
                    stat += "V"
                if len(stat) > 0:
                    spec = sid + " (" + stat + ")"
                else:
                    spec = sid
            else:
                spec = "not loaded"
            this["specID"] = spec

            cuts.append(this)
        table = hdtv.util.Table(cuts, params, sortBy="ID")

        if matrix.sym:
            sym = "symmetric"
        else:
            sym = "asymmetric"
        header = "\nmatrix ID = " + matrix.ID + ' "' + matrix.name + '" (' + sym + ")\n"
        footer = (
            "\n"
            + str(len(matrix.ids))
            + " cuts, "
            + str(count)
            + " loaded cut spectra."
        )

        table = hdtv.util.Table(
            cuts, params, sortBy="ID", extra_header=header, extra_footer=footer
        )
        return str(table)


class TvMatInterface(object):
    def __init__(self, matInterface):
        self.matIf = matInterface
        self.spectra = self.matIf.spectra

        prog = "matrix get"
        description = "load a matrix, i.e. the projections"
        description += "if the matrix is symmetric it only loads one projection"
        description += "if it is asymmetric both projections will be loaded."
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default=None,
            help="base id for loaded projections",
        )
        parser.add_argument("matrix_type", metavar="matrix-type", help="{asym,sym}")
        parser.add_argument(
            "filename", metavar="matrix-file", help="file with matrix to load"
        )
        hdtv.cmdline.AddCommand(
            prog, self.MatrixGet, level=0, fileargs=True, parser=parser
        )

        # FIXME
        prog = "matrix list"
        description = "list all loaded matrices and the belonging cuts and cut spectra."
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog)
        # TODO: sort for gates
        # parser.add_argument("-k", "--key-sort", action = "store", default = hdtv.options.Get("fit.list.sort_key"),
        #     help = "sort by key")
        # parser.add_argument("-r", "--reverse-sort", action = "store_true", default = False,
        #     help = "reverse the sort")
        parser.add_argument(
            "-m",
            "--matrix",
            action="store",
            default="all",
            help="select matrix to work on",
        )
        hdtv.cmdline.AddCommand(prog, self.MatrixList, parser=parser)

        # FIXME
        prog = "matrix project"
        description = "reload projection of matrix"

        prog = "matrix view"
        description = "show 2D view of the matrix"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog)
        parser.add_argument(
            "filename", metavar="matrix-file", help="file with matrix to load"
        )
        parser.add_argument("format", nargs="?", default=None, help="format of matrix")
        hdtv.cmdline.AddCommand(
            prog, self.MatrixView, level=0, fileargs=True, parser=parser
        )

        # FIXME
        prog = "matrix delete"
        description = "delete the matrix, with all cuts and cut spectra"

        prog = "cut marker"
        description = "set/delete a marker for cutting"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument("type", help="{background,region}")
        parser.add_argument("action", help="{delete,set}")
        parser.add_argument("position", type=float, help="position of marker")
        hdtv.cmdline.AddCommand(
            prog, self.CutMarkerChange, parser=parser, completer=self.MarkerCompleter
        )

        # FIXME: this should accept --cut to reload cut spectra for a cut
        prog = "cut execute"
        description = "execute cut"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        hdtv.cmdline.AddCommand(prog, self.CutExecute, level=0, parser=parser)

        prog = "cut clear"
        description = "clear cut marker and remove last cut if it was not stored"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        hdtv.cmdline.AddCommand(prog, self.CutClear, parser=parser)

        prog = "cut store"
        description = "deactivates last cut (overwrite protection)"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        hdtv.cmdline.AddCommand(prog, self.CutStore, parser=parser)

        prog = "cut activate"
        description = "reactivates a cut from the cutlist"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument("cutid", nargs="+", help="id of cut")
        hdtv.cmdline.AddCommand(prog, self.CutActivate, parser=parser)

        prog = "cut delete"
        description = "delete a cut (marker and spectrum)"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument("cutid", nargs="+", help="id of cut")
        hdtv.cmdline.AddCommand(prog, self.CutDelete, minargs=1, parser=parser)

        # FIXME
        prog = "cut show"
        description = "show a cut"

        # FIXME
        prog = "cut hide"
        description = "hide a cut"

        # List of MTViewers (2d matrix views)
        self.matviews = []

    def MatrixView(self, args):
        """
        Load a matrix from file, then display it in 2d
        """
        hist = SpecReader.GetMatrix(args.filename, args.format)

        title = hist.GetTitle()
        viewer = ROOT.HDTV.Display.MTViewer(400, 400, hist, title)
        self.matviews.append(viewer)

    def MatrixGet(self, args):
        """
        Load a matrix from file
        """
        if args.spectrum is not None:
            ID = args.spectrum
        else:
            ID = None
        if args.matrix_type == "sym":
            sym = True
        elif args.matrix_type == "asym":
            sym = False
        else:
            # FIXME: is there really no way to test that automatically????
            raise hdtv.cmdline.HDTVCommandError(
                "Please specify if matrix is of type asym or sym"
            )
        self.matIf.LoadMatrix(args.filename, sym, ID=ID)

    def MatrixList(self, args):
        """
        Show a overview of matrices with all cuts and cut spectra
        """
        # TODO: Parse ids
        ids = "ALL"
        matrices = set()

        for spec in self.spectra.dict.values():
            if hasattr(spec, "matrix") and spec.matrix is not None:
                if ids == "ALL" or spec.matrix.ID in ids:
                    matrices.add(spec.matrix)

        result = str()
        for mat in matrices:
            result += self.matIf.ListMatrix(mat)
        hdtv.ui.msg(html=result)

    def CutMarkerChange(self, args):
        """
        Set or delete a marker from command line
        """
        # complete markertype if needed
        mtype = self.MarkerCompleter(args.type)
        if len(mtype) == 0:
            raise hdtv.cmdline.HDTVCommandError(
                "Marker type %s is not valid" % args.type
            )
        action = self.MarkerCompleter(
            args.action,
            args=[
                args.action,
            ],
        )
        if len(action) == 0:
            raise hdtv.cmdline.HDTVCommandError("Invalid action: %s" % args.action)

        mtype = mtype[0].strip()
        # replace "background" with "cutbg" which is internally used
        if mtype == "background":
            mtype = "cutbg"
        # replace "region" with "cutregion" which is internally used
        if mtype == "region":
            mtype = "cutregion"
        action = action[0].strip()
        if action == "set":
            self.spectra.SetMarker(mtype, args.position)
        if action == "delete":
            self.spectra.RemoveMarker(mtype, args.position)

    def MarkerCompleter(self, text, args=[]):
        """
        Helper function for CutMarkerChange
        """
        if not args:
            mtypes = ["background", "region"]
            return hdtv.util.GetCompleteOptions(text, mtypes)
        elif len(args) == 1:
            actions = ["set", "delete"]
            return hdtv.util.GetCompleteOptions(text, actions)
        return []

    def CutExecute(self, args):
        return self.spectra.ExecuteCut()

    def CutClear(self, args):
        return self.spectra.ClearCut()

    def CutStore(self, args):
        return self.spectra.StoreCut()

    def CutActivate(self, args):
        """
        Activate a cut

        This marks a stored cut as active and copies its markers to the work cut
        """
        spec = self.spectra.GetActiveObject()
        if spec is None:
            hdtv.ui.warning("No active spectrum")
            return
        if not hasattr(spec, "matrix") or spec.matrix is None:
            hdtv.ui.warning("Active spectrum does not belong to a matrix")
            return
        ids = hdtv.util.ID.ParseIds(args.cutid, spec.matrix)
        if len(ids) == 1:
            hdtv.ui.msg("Activating cut %s" % ids[0])
            self.spectra.ActivateCut(ids[0])
        elif len(ids) == 0:
            hdtv.ui.msg("Deactivating cut")
            self.spectra.ActivateCut(None)
        else:
            raise hdtv.cmdline.HDTVCommandError("Can only activate one cut")

    def CutDelete(self, args):
        """
        Delete a cut and its cut spectrum
        """
        spec = self.spectra.GetActiveObject()
        if spec is None:
            hdtv.ui.warning("No active spectrum")
            return
        if not hasattr(spec, "matrix") or spec.matrix is None:
            hdtv.ui.warning("Active spectrum does not belong to a matrix")
            return
        ids = hdtv.util.ID.ParseIds(args.cutid, spec.matrix)
        for ID in ids:
            cut = spec.matrix.Pop(ID)
            # delete also cut spectrum
            if cut.spec is not None:
                sid = self.spectra.Index(cut.spec)
                hdtv.ui.msg("Remove spec %s" % sid)
                self.spectra.Pop(sid)


# plugin initialisation
import __main__

matrix_interface = MatInterface(__main__.spectra)
hdtv.cmdline.RegisterInteractive("mat", matrix_interface)



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import os

import pytest

from tests.helpers.utils import setup_io, redirect_stdout, isclose

from hdtv.util import monkey_patch_ui, strip_tags

monkey_patch_ui()

import __main__

import hdtv.session

try:
    __main__.spectra = hdtv.session.Session()
except RuntimeError:
    pass
spectra = __main__.spectra

from hdtv.plugins.specInterface import spec_interface
from hdtv.plugins.fitInterface import fit_interface
import hdtv.plugins.fitlist

testspectrum = os.path.join(os.path.curdir, "tests", "share", "osiris_bg.spc")


@pytest.fixture(autouse=True)
def prepare():
    fit_interface.ResetFitterParameters()
    hdtv.options.Set("table", "classic")
    hdtv.options.Set("uncertainties", "short")
    spec_interface.LoadSpectra(testspectrum)
    for i in range(5):
        spec_interface.LoadSpectra(testspectrum)
        spectra.Get(str(i)).cal = [0, (i + 1) * 0.5]
    yield
    spectra.Clear()


def get_list(no_err=True):
    f, ferr = setup_io(2)
    with redirect_stdout(f, ferr):
        spec_interface.ListSpectra()
    if no_err:
        assert ferr.getvalue().strip() == ""
        return strip_tags(f.getvalue())
    else:
        return strip_tags(f.getvalue()), ferr.getvalue()


def test_spectra_loaded():
    res = get_list()
    for i in range(5):
        assert str(i) + " |    V | osiris_bg.spc |    0" in res
    assert "5 |   AV | osiris_bg.spc |    0" in res


def test_spectra_show_first():
    spectra.ShowFirst()
    res = get_list()
    assert "0 |   AV | osiris_bg.spc |    0" in res
    for i in range(1, 6):
        assert str(i) + " |      | osiris_bg.spc |    0" in res


def test_spectra_show_next():
    spectra.ShowFirst()
    for i in range(1, 6):
        spectra.ShowNext()
        res = get_list()
        for j in range(6):
            if j == i:
                assert str(j) + " |   AV | osiris_bg.spc |    0" in res
            else:
                assert str(j) + " |      | osiris_bg.spc |    0" in res


def test_spectra_show_last():
    spectra.ShowLast()
    res = get_list()
    for j in range(6):
        if j == 5:
            assert str(j) + " |   AV | osiris_bg.spc |    0" in res
        else:
            assert str(j) + " |      | osiris_bg.spc |    0" in res


def test_spectra_show_prev():
    spectra.ShowLast()
    for i in range(4, 0, -1):
        spectra.ShowPrev()
        res = get_list()
        for j in range(6):
            if j == i:
                assert str(j) + " |   AV | osiris_bg.spc |    0" in res
            else:
                assert str(j) + " |      | osiris_bg.spc |    0" in res


def test_spectra_show_number():
    for i in range(6):
        spectra.ShowObjects(hdtv.util.ID.ParseIds(str(i), __main__.spectra))
        res = get_list()
        for j in range(6):
            if j == i:
                assert str(j) + " |   AV | osiris_bg.spc |    0" in res
            else:
                assert str(j) + " |      | osiris_bg.spc |    0" in res


def test_spectra_show_all():
    spectra.ShowLast()
    spectra.ShowAll()
    res = get_list()
    for j in range(6):
        if j == 5:
            assert str(j) + " |   AV | osiris_bg.spc |    0" in res
        else:
            assert str(j) + " |    V | osiris_bg.spc |    0" in res


def test_spectra_activate_number():
    spectra.ShowAll()
    for i in range(6):
        spectra.ActivateObject(str(i))
        res = get_list()
        for j in range(6):
            if j == i:
                assert str(j) + " |   AV | osiris_bg.spc |    0" in res
            else:
                assert str(j) + " |    V | osiris_bg.spc |    0" in res


def test_spectra_show_active():
    spectra.ShowAll()
    spectra.ActivateObject(str(5))
    spectra.ShowObjects(spectra.activeID)
    res = get_list()
    for j in range(6):
        if j == 5:
            assert str(j) + " |   AV | osiris_bg.spc |    0" in res
        else:
            assert str(j) + " |      | osiris_bg.spc |    0" in res


def test_spectra_activate_hidden():
    spectra.ShowAll()
    spectra.ActivateObject(str(5))
    spectra.ShowObjects(spectra.activeID)
    spectra.ActivateObject("4")
    res = get_list()
    for j in range(6):
        if j == 5:
            assert str(j) + " |    V | osiris_bg.spc |    0" in res
        elif j == 4:
            assert str(j) + " |   AV | osiris_bg.spc |    0" in res
        else:
            assert str(j) + " |      | osiris_bg.spc |    0" in res


def test_spectra_remove_one():
    spectra.Pop("3")
    res = get_list()
    assert not "3 | " in res


def test_spectra_reload_one():
    spectra.Pop("3")
    spec_interface.LoadSpectra(testspectrum)
    res = get_list()
    assert "3 |   AV" in res


def test_spectra_show_next_overflow():
    spectra.ShowObjects(hdtv.util.ID.ParseIds(str(3), __main__.spectra))
    for i in range(3, 10):
        spectra.ShowNext()
        res = get_list()
        for j in range(6):
            if j == (i + 1) % 6:
                assert str(j) + " |   AV | osiris_bg.spc |    0" in res
            else:
                assert str(j) + " |      | osiris_bg.spc |    0" in res



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2011-2013  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import os
import numpy
import scipy
import matplotlib

matplotlib.use("agg")  # Must be before import pylab!
import pylab

import matplotlib.transforms as transforms

import hdtv.ui
import hdtv.cmdline
import hdtv.cal
import hdtv.color


# TODO: add cut marker


class PrintOut(object):
    def __init__(self, spectra, energies=False):
        self.spectra = spectra
        self.labels = energies

    def Execute(self):
        """
        creates a plot with all currently visible objects
        """
        # extract visible spectrum objects
        ids = list(self.spectra.visible)
        specs = [self.spectra.dict[i] for i in ids]

        # create matplotlib plot
        pylab.rc("font", size=16)
        pylab.rc("lines", linewidth=1)
        pylab.rc("svg", fonttype="path")

        pylab.figure(figsize=(8.5, 4.2))

        # add spectra, fits, marker etc.
        for spec in specs:
            self.PrintHistogram(spec)
            # extract visible fits for this spectrum
            fids = list(spec.visible)
            fits = [spec.dict[i] for i in fids]
            # add workFit
            fits.append(self.spectra.workFit)
            for fit in fits:
                if self.spectra.window.IsInVisibleRegion(fit, part=True):
                    self.PrintFit(fit)

        # viewport limits
        x1 = self.spectra.viewport.GetXOffset()
        x2 = x1 + self.spectra.viewport.GetXVisibleRegion()
        y1 = self.spectra.viewport.GetYOffset()
        y2 = y1 + self.spectra.viewport.GetYVisibleRegion()

        # apply limits to plot
        pylab.xlim(x1, x2)
        pylab.ylim(y1, y2)

    def PrintHistogram(self, spec):
        """
        print histogramm
        """
        nbins = spec.hist.hist.GetNbinsX()
        # create values for x axis
        en = numpy.arange(nbins)
        # shift in order to get values at bin center
        en = en - 0.5
        # calibrate
        en = self.ApplyCalibration(en, spec.cal)
        # extract bin contents to numpy array
        data = numpy.zeros(nbins)
        for i in range(nbins):
            data[i] = spec.hist.hist.GetBinContent(i)
        # create spectrum plot
        (r, g, b) = hdtv.color.GetRGB(spec.color)
        pylab.step(en, data, color=(r, g, b), label=spec.name)

    def PrintFit(self, fit):
        """
        print fit including peak marker and functions
        """
        # peak marker
        for p in fit.peakMarkers:
            self.PrintMarker(p)
        # print all marker for a workFit
        if fit.active:
            for m in fit.bgMarkers:
                self.PrintMarker(m)
            for m in fit.regionMarkers:
                self.PrintMarker(m)
        # peak function
        func = fit.dispPeakFunc
        if fit.active:
            color = hdtv.color.region
        else:
            color = fit.color
        if func:
            self.PrintFunc(func, fit.cal, color)
        # background function
        bgfunc = fit.dispBgFunc
        if fit.active:
            color = hdtv.color.bg
        else:
            color = fit.color
        if bgfunc:
            self.PrintFunc(bgfunc, fit.cal, fit.color)
        i = 0
        for p in fit.peaks:
            # maybe energy labels for each peak
            if self.labels:
                self.PrintLabel(p, i)
                i = i + 1
            # maybe functions for each peak
            if fit._showDecomp:
                peakfunc = p.displayObj
                self.PrintFunc(peakfunc, p.cal, p.color)

    def PrintMarker(self, marker):
        """
        print marker
        """
        if marker.active:
            color = marker._activeColor
        else:
            color = marker.color
        (r, g, b) = hdtv.color.GetRGB(color)
        pos = marker.p1.pos_cal
        pylab.axvline(pos, color=(r, g, b))
        if marker.p2 is not None:
            pos = marker.p2.pos_cal
            pylab.axvline(pos, color=(r, g, b))

    def PrintLabel(self, peak, i):
        """
        print energy labels for peaks
        """
        x = peak.pos_cal.value
        ax = pylab.gca()
        trans = transforms.blended_transform_factory(ax.transData, ax.transAxes)
        pylab.text(
            x,
            0.94 - 0.06 * i,
            "{:S}".format(peak.pos_cal),
            transform=trans,
            size="small",
        )

    def PrintFunc(self, func, cal, color):
        """
        print function by evaluation it at several points
        """
        x1 = func.GetMinCh()
        x2 = func.GetMaxCh()
        nbins = int(x2 - x1) * 20
        en = numpy.linspace(x1, x2, nbins)
        data = numpy.zeros(nbins)
        for i in range(nbins):
            data[i] = func.Eval(en[i])
        en = self.ApplyCalibration(en, cal)
        # create function plot
        (r, g, b) = hdtv.color.GetRGB(color)
        pylab.plot(en, data, color=(r, g, b))

    def ApplyCalibration(self, en, cal):
        """
        return calibrated values
        """
        if cal and not cal.IsTrivial():
            calpolynom = hdtv.cal.GetCoeffs(cal)
            # scipy needs highes order first
            calpolynom.reverse()
            en = scipy.polyval(calpolynom, en)
        return en


class PrintInterface(object):
    def __init__(self, spectra):
        self.spectra = spectra

        # command line interface
        prog = "print"
        description = "Prints all visible items to file. The file format is specified by the filename extension."
        description += (
            "Supported formats are: emf, eps, pdf, png, ps, raw, rgba, svg, svgz. "
        )
        description += "If no filename is given, an interactive mode is entered and the plot can be manipulated using pylab."
        description += (
            "Change to the python prompt and import the pylab module for that to work."
        )
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-F",
            "--force",
            action="store_true",
            default=False,
            help="overwrite existing files without asking",
        )
        parser.add_argument(
            "-y", "--ylabel", action="store", default=None, help="add label for y-axis"
        )
        parser.add_argument(
            "-x", "--xlabel", action="store", default=None, help="add label for x-axis"
        )
        parser.add_argument(
            "-t", "--title", action="store", default=None, help="add title for plot"
        )
        parser.add_argument(
            "-l", "--legend", action="store_true", default=False, help="add legend"
        )
        parser.add_argument(
            "-e",
            "--energies",
            action="store_true",
            default=False,
            help="add energy labels to each fitted peak",
        )
        parser.add_argument(
            "filename",
            metavar="output-file",
            nargs="?",
            default=None,
            help="file to print to",
        )
        hdtv.cmdline.AddCommand(prog, self.Print, fileargs=True, parser=parser)

    def Print(self, args):
        pylab.ioff()

        p = PrintOut(self.spectra, args.energies)
        p.Execute()

        # add some decorations
        if args.title:
            pylab.title(args.title)
        if args.ylabel:
            pylab.ylabel(args.ylabel)
        if args.xlabel:
            pylab.xlabel(args.xlabel)
        if args.legend and len(self.spectra) > 0:
            legend = pylab.legend(prop=dict(size="x-small"))
            legend.draw_frame(False)

        # save finished plot to file
        if args.filename:
            fname = hdtv.util.user_save_file(args.filename, args.force)
            if not fname:
                return
            pylab.ioff()
            pylab.savefig(fname, bbox_inches="tight")
        else:
            # else go to interactive mode
            pylab.ion()
            # hack to open the plot window
            pylab.text(0, 0, "")


# plugin initialisation
import __main__

print_interface = PrintInterface(__main__.spectra)
hdtv.cmdline.RegisterInteractive("p", print_interface)
hdtv.ui.debug("Loaded user interface for printing")



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Plugin for executing a python script from the command line
"""

import os
import sys

import hdtv.ui
import hdtv.cmdline


def run(args):
    """
    Executes a python script from the hdtv command line (via execfile)
    """
    fname = os.path.expanduser(args[0])
    try:
        with open(fname) as f:
            hdtv.ui.msg("Running script %s" % fname)
            exec(compile(f.read(), fname, "exec"))
            hdtv.ui.msg("Finished!")
    except IOError as err:
        hdtv.ui.error(str(err))


hdtv.cmdline.AddCommand("run", run, nargs=1, fileargs=True)

hdtv.ui.debug("Loaded run plugin")



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import sys
import os
import contextlib

try:  # Python2
    from StringIO import StringIO
except ImportError:  # Python3
    from io import StringIO

from hdtv.ui import ui
import hdtv.cmdline


def setup_io(num=1):
    """
    Setup several StringIOs that can be used for stdout redirection.
    """
    return [StringIO() for _ in range(num)]


@contextlib.contextmanager
def redirect_stdout(target_out, target_err=None, target_debug=None):
    original_out = sys.stdout
    original_out_hdtv = ui.stdout
    sys.stdout = target_out
    ui.stdout = target_out
    if target_err:
        original_err = sys.stderr
        sys.stderr = target_err
        original_err_hdtv = ui.stderr
        ui.stderr = target_err
    if target_debug:
        original_debug_hdtv = ui.debugout
        ui.debugout = target_debug
    yield
    sys.stdout = original_out
    ui.stdout = original_out_hdtv
    if target_err:
        sys.stderr = original_err
        ui.stderr = original_err_hdtv
    if target_debug:
        ui.debugout = original_debug_hdtv


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    """
    Determine if two floats are close to each other.
    """
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def hdtvcmd(*commands):
    """
    Execute hdtv command(s) and return stdout and stderr output.
    """
    f, ferr = setup_io(2)
    with redirect_stdout(f, ferr):
        for command in commands:
            hdtv.cmdline.command_line.DoLine(command)
    return f.getvalue().strip(), ferr.getvalue().strip()



# -*- coding: utf-8 -*-
# PGAA database from Institute of Isotopes, Hungarian Academey of Science,
# Budapest

from uncertainties import ufloat, ufloat_fromstr

from hdtv.database.common import *
import hdtv.cmdline
import hdtv.ui


class PGAAGamma(Gamma):
    """
    Store data about PGAA gammas

    k0_comp is the comparator element for k0 values which k0 should be normed to 1.0
        given as tuple (Z, A) (default: 1-H)
    """

    k0_norm = None
    __slots__ = ("ID", "nuclide", "energy", "sigma", "intensity", "halflife", "_k0")

    def __init__(
        self,
        nuclide,
        energy,
        sigma=None,
        intensity=None,
        k0=None,
        halflife=None,
        k0_comp=(1, 1),
    ):

        super(PGAAGamma, self).__init__(nuclide, energy, sigma, intensity)
        self.halflife = halflife
        self._k0 = k0  # TODO
        if nuclide == Nuclides(k0_comp[0], k0_comp[1])[0] and not PGAAGamma.k0_norm:
            # Normalize reference element to k0=1.0
            PGAAGamma.k0_norm = 1.0 / self.getk0(isNorm=True)

    def getk0(self, isNorm=False):

        if self._k0 is None:  # k0 was not given: we have to calculate
            if isNorm:
                k0 = self.sigma / self.nuclide.element.m
            else:
                try:
                    k0 = (self.sigma / self.nuclide.element.m) * PGAAGamma.k0_norm
                except TypeError:
                    k0 = None
            return k0
        else:
            return self._k0

    k0 = property(getk0)

    def __str__(self):

        text = str()
        text += super(PGAAGamma, self).__str__()
        text += "k0:        " + str(self.k0) + "\n"
        return text


class PGAAlib_IKI2000(GammaLib):
    """
    PGAA library of the Institute of Isotopes, Hungarian Academy of Sciences, Budapest
    """

    def __init__(
        self,
        csvfile=os.path.join(hdtv.datadir, "PGAAlib-IKI2000.dat"),
        has_header=True,
        k0_comp=(1, 1),
    ):

        super(PGAAlib_IKI2000, self).__init__()

        # Header for table printout
        self.fOrderedHeader = [
            "Z",
            "A",
            "El",
            "Energy/(keV)",
            "Intensity",
            "Sigma/(b)",
            "k0",
            "Halflife/(s)",
        ]
        # Conversion functions for parameter
        self.fParamConv = {
            "z": int,
            "a": int,
            "symbol": str,
            "energy": float,
            "intensity": float,
            "sigma": float,
            "k0": float,
            "halflife": float,
        }
        self.name = "PGAAlib_IKI2000"
        self.description = (
            "PGAA database, Inst. of Isotopes, Hungarian Academey of Science"
        )
        self.csvfile = csvfile
        self._has_header = has_header
        self.k0_comp = k0_comp

    def open(self):

        if self.opened:
            return True

        try:
            datfile = open(self.csvfile, "r", encoding="utf-8")
        except TypeError:
            datfile = open(self.csvfile, "rb")

        reader = csv.reader(datfile)

        if self._has_header:
            next(reader)

        try:
            for line in reader:
                Z = int(line[0])
                A = int(line[1])
                energy = ufloat(float(line[2]), float(line[3]))
                sigma = ufloat(float(line[4]), float(line[5]))
                intensity = ufloat(float(line[6]), 0) / 100.0
                try:
                    halflife = ufloat(float(line[7]), 0)
                except ValueError:
                    halflife = None
                gamma = PGAAGamma(
                    Nuclides(Z, A)[0],
                    energy,
                    sigma=sigma,
                    intensity=intensity,
                    halflife=halflife,
                    k0_comp=self.k0_comp,
                )
                self.append(gamma)
        except csv.Error as e:
            raise HDTVCommandAbort(
                "file %s, line %d: %s" % (self.csvfile, reader.line_num, e)
            )
        else:
            self.opened = True
        finally:
            datfile.close()


class PromptGammas(GammaLib):
    """
    Extensive IAEA Prompt-Gamma library
    """

    def __init__(
        self,
        csvfile=os.path.join(hdtv.datadir, "PromptGammas.dat"),
        has_header=True,
        k0_comp=(1, 1),
    ):

        super(PromptGammas, self).__init__()

        # Header for table printout
        self.fOrderedHeader = ["Z", "A", "El", "Energy/(keV)", "Sigma/(b)", "k0"]
        # Conversion functions for parameter
        self.fParamConv = {
            "z": int,
            "a": int,
            "symbol": str,
            "energy": float,
            "sigma": float,
            "k0": float,
        }
        self.name = "PromptGammas"
        self.description = "Extensive Prompt-Gamma library"
        self.csvfile = csvfile
        self._has_header = has_header
        self.k0_comp = k0_comp

    def open(self):

        if self.opened:
            return True

        try:
            datfile = open(self.csvfile, "r", encoding="utf-8")
        except TypeError:
            datfile = open(self.csvfile, "rb")

        reader = csv.reader(datfile)

        if self._has_header:
            next(reader)

        try:
            for line in reader:
                A = int(line[0])
                Z = int(line[1])
                energy = ufloat_fromstr(line[2])
                sigma = ufloat_fromstr(line[3])
                k0 = ufloat_fromstr(line[4])
                gamma = PGAAGamma(
                    Nuclides(Z, A)[0], energy, sigma=sigma, k0=k0, k0_comp=self.k0_comp
                )
                self.append(gamma)
        except csv.Error as e:
            hdtv.ui.error("file %s, line %d: %s" % (self.csvfile, reader.line_num, e))
        else:
            self.opened = True
        finally:
            datfile.close()



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Helper functions for access to ROOT files, treating them like POSIX
directories
"""

import os
import fnmatch
import ROOT
from prompt_toolkit.completion import Completer, Completion
from hdtv.util import SplitCmdlines

# Required for Get() to work (otherwise, histogram objects are automatically
# destroyed when the file containing them is closed).
ROOT.TH1.AddDirectory(False)


class Quiet:
    def __init__(self, level=ROOT.kInfo + 1):
        """
        Context manager to suppress ROOT output
        """
        self.level = level

    def __enter__(self):
        self.oldlevel = ROOT.gErrorIgnoreLevel
        ROOT.gErrorIgnoreLevel = self.level

    def __exit__(self, *a):
        ROOT.gErrorIgnoreLevel = self.oldlevel


class RootCdCompleter(Completer):
    def get_completions(self, document, complete_event):
        yield from GetSuggestions(document, True)


class RootCompleter(Completer):
    def get_completions(self, document, complete_event):
        yield from GetSuggestions(document, False)


def GetSuggestions(document, dirs_only=False):
    word_before_cursor = document.get_word_before_cursor()
    cmds, _ = SplitCmdlines(document.text_before_cursor)

    if word_before_cursor:
        last_path = cmds[-1][-1]
    else:
        last_path = ""

    if last_path:
        (_, word_before_cursor) = os.path.split(last_path)

    default_style = "fg:#000000"
    default_selected_style = "bg:white fg:ansired"

    for opt in CombinedSuggestions(last_path):
        yield Completion(
            opt,
            -len(word_before_cursor),
            style=default_style,
            selected_style=default_selected_style,
        )


def CombinedSuggestions(text, dirs_only=False):
    if ".root" in text:
        # User wants too look into unopened ROOT File
        prev_root_dir = ROOT.gDirectory.GetDirectory("")
        try:
            rootfilebase, pattern = text.rsplit(".root", 1)
            rootfilename = rootfilebase + ".root"
            if IsROOTFile(rootfilename):
                with Quiet(ROOT.kError):
                    rf = ROOT.TFile.Open(rootfilename)
                    options = RootSuggestions(pattern, dirs_only)
                    rf.Close()
                return options
        except Exception:
            return []
        finally:
            # Restore the saved ROOT directory
            if prev_root_dir:
                prev_root_dir.cd()
    else:
        options = []
        # Suggestions for the currently open ROOT file
        options += RootSuggestions(text, dirs_only)
        # Suggestions from the file system
        options += FileSuggestions(text)
        return options


def RootSuggestions(text, dirs_only=False):
    # Suggest completions from a ROOT directory
    prev_root_dir = ROOT.gDirectory.GetDirectory("")
    options = []

    try:
        # Descend to Subdirectories if required
        path_segments = text.rsplit("/", 1)
        if len(path_segments) == 2:
            # TODO Handle case where path_segments[0]
            # does not exist or is no directory
            with Quiet(ROOT.kError + 1):
                ROOT.gDirectory.cd(path_segments[0])
            pattern = path_segments[1]
        else:
            pattern = path_segments[0]

        # Get Contents of ROOT File
        keys = ROOT.gDirectory.GetListOfKeys() or []
        for k in keys:
            name = k.GetName()
            if name.startswith(pattern):
                if k.GetClassName() == "TDirectoryFile":
                    options.append(name + "/")
                elif not dirs_only:
                    options.append(name + " ")
    finally:
        # Restore the saved ROOT directory
        if prev_root_dir:
            prev_root_dir.cd()

    return sorted(options)


def FileSuggestions(text):
    # Suggest directories and ROOT Files
    try:
        options = []
        # Descend to Subdirectories if required
        path_segments = text.rsplit("/", 1)
        if len(path_segments) == 2:
            path = path_segments[0]
            pattern = path_segments[1]
        else:
            path = "."
            pattern = path_segments[0]

        for name in os.listdir(path or "."):
            if name.startswith(pattern):
                full_path = os.path.join(path, name)
                if os.path.isdir(full_path) or IsROOTFile(full_path):
                    options.append(name + "/")
        return sorted(options)
    except Exception:
        return []


def GetRelDirectory(cur_posix_path, cur_root_dir, path):
    """
    Resolve the relative path path with respect to the current directory
    specified by cur_posix_path and cur_root_dir.

    cur_posix_path is a string specifying the current working directory. If
    cur_root_dir is not None, it is taken as a TDirectoryFile object specifying
    the working directory inside the ROOT file. The ROOT file is expected to
    reside inside the directory specified by cur_posix_path.

    Returns a triple (new_posix_path, new_root_file, new_root_dir).
    new_posix_path is the new posix path.
    new_root_file is the ROOT file in which new_root_dir resides, or None if
    new_root_dir resides in the same file as cur_root_dir. If new_root_file is
    not None, the caller is expected to close the file after being done with it.
    (Note: a path moving out of a ROOT file and then back in
    is considered to refer to a different file, with the consequences mentioned
    above.)
    new_root_dir is the new ROOT TDirectory.

    If path is invalid, (None, None, None) is returned.
    """
    # Save the current ROOT directory

    # Annoyingly, ROOT automatically changes the ''current directory''
    # (ROOT.gDirectory) to that file if a ROOT file is opened. Obviously,
    # completion should *not* change the current directory, but we need to open
    # the file to see what is inside. Thus, we have to save the current
    # directory and restore it after we are done. Note that this trick is *not*
    # thread-safe!

    # Saving the current directory turns out to be tricky; see
    # http://root.cern.ch/phpBB2/viewtopic.php?t=6382

    prev_root_dir = ROOT.gDirectory.GetDirectory("")

    try:
        error = False
        rfile = None
        pcomp = path.split("/")

        # Handle absolute paths
        if len(pcomp) >= 2 and pcomp[0] == "":
            del pcomp[0]
            cur_posix_path = "/"
            cur_root_dir = None

        # Descend down the path
        for pc in pcomp:
            if cur_root_dir is not None:
                if pc == "" or pc == ".":
                    pass
                elif pc == "..":
                    cur_root_dir = cur_root_dir.GetMotherDir()
                    if cur_root_dir is None:
                        if rfile is not None:
                            # print "Info: closing %s" % str(rfile)
                            rfile.Close()
                else:
                    cur_root_dir = cur_root_dir.Get(pc)
                    if not isinstance(cur_root_dir, ROOT.TDirectoryFile):
                        error = True
                        break
            else:
                full_name = os.path.join(cur_posix_path, pc)
                if os.path.isdir(full_name):
                    cur_posix_path = full_name
                elif IsROOTFile(full_name):
                    # print("Info: Opening %s" % full_name)
                    # Disable warnings when opening files
                    ROOT.gErrorIgnoreLevel = ROOT.kError
                    rfile = ROOT.TFile(full_name)
                    ROOT.gErrorIgnoreLevel = ROOT.kInfo
                    if rfile.IsZombie():
                        error = True
                        break
                    cur_root_dir = rfile
                else:
                    error = True
                    break

        if error:
            if rfile:
                # print "Info: closing %s" % str(rfile)
                rfile.Close()
            return (None, None, None)

        return (cur_posix_path, rfile, cur_root_dir)

    finally:
        # Restore the saved ROOT directory
        if prev_root_dir:
            prev_root_dir.cd()


def Get(cur_posix_path, cur_root_dir, pattern):
    """
    Function to load objects from ROOT files, treating them like POSIX
    directories.

    cur_posix_path specifies the current working directory.
    If cur_root_dir is a ROOT TDirectoryFile object, searching starts relative
    to that directory. If is is None, searching starts relative to
    cur_posix_path.    Pattern is a shell-style pattern, e.g. "*.root/spec/*".

    Returns a list of ROOT objects.
    """
    pcomp = pattern.split("/")

    if pcomp[0] == "":
        del pcomp[0]
        cur_root_dir = None
        cur_posix_path = "/"

    # Save the current ROOT directory
    prev_root_dir = ROOT.gDirectory.GetDirectory("")

    if cur_root_dir is None:
        objs = RecursivePathMatch(cur_posix_path, pcomp)
    else:
        objs = RecursiveROOTMatch(cur_posix_path, cur_root_dir, pcomp)

    # Restore the saved ROOT directory
    if prev_root_dir:
        prev_root_dir.cd()

    return objs


def RecursivePathMatch(cur_path, pcomp):
    if pcomp[0] in (".", ".."):
        if len(pcomp) > 1:
            return RecursivePathMatch(os.path.join(cur_path, pcomp[0]), pcomp[1:])
        else:
            return []

    matched_objects = []
    for name in os.listdir(cur_path):
        if fnmatch.fnmatch(name, pcomp[0]):
            full_name = os.path.join(cur_path, name)
            if os.path.isfile(full_name) and IsROOTFile(full_name):
                rfile = ROOT.TFile(full_name)
                matched_objects += RecursiveROOTMatch(cur_path, rfile, pcomp[1:])
                rfile.Close()
            elif os.path.isdir(full_name):
                matched_objects += RecursivePathMatch(full_name, pcomp[1:])
    return matched_objects


def RecursiveROOTMatch(rfile_path, rfile_dir, pcomp):
    """
    Recursively descend into ROOT file directory structure.
    rfile_path is a string specifying the POSIX directory of the ROOT file
    rfile_dir is a TDirectoryFile object specifying the current directory
     in the ROOT file
    """
    if pcomp[0] == ".":
        if len(pcomp) <= 1:
            return []
        else:
            return RecursiveROOTMatch(rfile_path, rfile_dir, pcomp[1:])
    elif pcomp[0] == "..":
        if len(pcomp) <= 1:
            return []
        else:
            mother_dir = rfile_dir.GetMotherDir()
            if mother_dir is not None:
                return RecursiveROOTMatch(rfile_path, mother_dir, pcomp[1:])
            else:
                return RecursivePathMatch(rfile_path, pcomp[1:])
    else:
        keys = rfile_dir.GetListOfKeys()
        if keys is None:
            return []

        matched_objects = []
        for k in keys:
            name = k.GetName()
            if fnmatch.fnmatch(name, pcomp[0]):
                if len(pcomp) == 1:
                    matched_objects.append(k.ReadObj())
                elif k.GetClassName() == "TDirectoryFile":
                    matched_objects += RecursiveROOTMatch(
                        rfile_path, k.ReadObj(), pcomp[1:]
                    )

    return matched_objects


def IsROOTFile(fname):
    """
    Tests if fname is a ROOT file
    """
    # Unfortunately, ROOT does not seem to provide a way to check if a given
    # file is a ROOT file, except for actually opening it. (TBrowser seems to
    # go by the filename extension.)

    try:
        if not os.path.isfile(fname):
            return False
        with open(fname, "rb") as f:
            ident = f.read(4).decode()
            return ident == "root"
    except Exception:
        return False



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import re
import sys

import pytest

from tests.helpers.utils import redirect_stdout, hdtvcmd

import hdtv.cmdline
import hdtv.options
import hdtv.plugins.dblookup


@pytest.fixture(autouse=True)
def prepare():
    hdtv.options.Set("table", "classic")
    hdtv.options.Set("uncertainties", "short")
    hdtv.options.Set("database.db", "pgaalib_iki2000")
    yield


def test_cmd_db_info():
    f, ferr = hdtvcmd("db info")
    assert "Database" in f
    assert "Valid fields" in f
    assert ferr == ""


def test_cmd_db_list():
    f, ferr = hdtvcmd("db list")
    assert "promptgammas" in f
    assert "pgaalib_iki2000" in f
    assert ferr == ""


def test_cmd_db_lookup_base():
    f, ferr = hdtvcmd("db lookup")
    assert "usage" in f
    assert "arguments" in ferr

    f, ferr = hdtvcmd("db lookup 0")
    assert ferr == ""
    assert "Found 0 results" in f


@pytest.mark.parametrize("specs", ["Intensity=0.1", "Sigma=0.03", "Energy=510"])
def test_cmd_db_lookup_specs(specs):
    assert count_results("db lookup {}".format(specs)) > 0


@pytest.mark.parametrize("specs", ["Intensity=0.3", "Sigma=0.04", "Energy=139.940"])
def test_cmd_db_lookup_fuzziness(specs):
    results_narrow = count_results("db lookup {} -f 0.005".format(specs))
    results_broad = count_results("db lookup {} --fuzziness 0.5".format(specs))
    assert results_narrow > 0
    assert results_broad > 0
    assert results_broad > results_narrow


@pytest.mark.parametrize("specs", ["k0=2", "510", "a=40", "z=33"])
def test_cmd_db_lookup_sort_reverse(specs):
    test_cmd_db_lookup_sort_key(specs, reverse=True)


@pytest.mark.parametrize("specs", ["k0=3", "511", "a=20", "z=10"])
def test_cmd_db_lookup_sort_key(specs, reverse=False):
    if reverse:
        f, ferr = hdtvcmd("db lookup {} -k energy -r".format(specs))
    else:
        f, ferr = hdtvcmd("db lookup {} -k energy".format(specs))
    old_value = sys.float_info.max if reverse else sys.float_info.min
    for line in f.split("\n")[2:-3]:
        new_value = float(line.split("|")[3].split("(")[0].strip())
        if reverse:
            assert new_value < old_value
        else:
            assert new_value > old_value
        old_value = new_value


@pytest.mark.parametrize("db", ["promptgammas", "pgaalib_iki2000"])
def test_cmd_db_set(db):
    f, ferr = hdtvcmd("db set {}".format(db))
    assert "loaded" in f
    assert hdtv.options.Get("database.db") == db


def count_results(query):
    f, ferr = hdtvcmd(query)
    return int(re.search(r"Found (\d+) results", f).groups()[0])



#!/usr/bin/env python

import struct
import os

f = open("mat.mtx", "wb")

for y in range(0, 1024):
    for x in range(0, 1024):
        z = 1
        if x >= 200 and y >= 200:
            z = 2

        f.write(struct.pack("l", z))

f.close()

os.system("/ikp/bin/matconv -I 1k.1k.le4 -D mat.mtx")
os.system("/ikp/bin/matproj mat.mtx")
